from __future__ import annotations 
from dataclasses import dataclass, field, InitVar
import numpy as np
from pyvol_terminal import utils
from typing import Any, List
import copy
from pyvol_terminal.plotting_engines import MetricEngine
from typing import Dict, ClassVar, Optional
from pyvol_terminal import exceptions
import warnings

def custom_showwarning(message, category, filename, lineno, file=None, line=None):
    print(f"UserWarning: {message} ({filename}, line {lineno})")
warnings.showwarning = custom_showwarning


@dataclass(slots=True)
class BaseData:
    n_instruments: int
    name_index_map: Dict[str, int]
    price: np.ndarray
    
        
@dataclass(slots=True)
class RawFutures(BaseData):
    OTM: np.ndarray
    

@dataclass(slots=True)
class RawOptions(BaseData):
    ivol: np.ndarray
    delta: np.ndarray
    delta_mag: np.ndarray
    gamma: np.ndarray
    vega: np.ndarray
    theta: np.ndarray
    rho: np.ndarray
    moneyness: np.ndarray
    log_moneyness: np.ndarray
    standardised_moneyness: np.ndarray
    OTM: np.ndarray
    call_flag: np.ndarray
    valid_price: np.ndarray
    underlying_px: np.ndarray

    def update_IVOL(self, idx, implied_volatility):
        self.ivol[idx] = implied_volatility
    
    def update_all_metrics_name(self, ticker, price, ivol, delta, delta_mag, gamma, vega, theta, rho, moneyness,
                                log_moneyness, standardised_moneyness, OTM, call_flag, underlying_px, valid_price):
        idx = self.name_index_map[ticker]
        self.price[idx] = price
        self.ivol[idx] = ivol
        self.delta[idx] = delta
        self.delta_mag[idx] = delta_mag
        self.gamma[idx] = gamma
        self.vega[idx] = vega
        self.theta[idx] = theta
        self.rho[idx] = rho
        self.moneyness[idx] = moneyness
        self.log_moneyness[idx] = log_moneyness
        self.standardised_moneyness[idx] = standardised_moneyness
        self.OTM[idx] = OTM
        self.call_flag[idx] = call_flag
        self.valid_price[idx] = valid_price
        self.underlying_px[idx] = underlying_px
    
    def update_all_metrics(self, idx, jdx, ivol, delta, delta_mag, gamma, vega, moneyness,
                           log_moneyness, standardised_moneyness, OTM, valid_price, underlying_px):
        self.ivol[idx] = ivol[jdx]
        self.delta[idx] = delta[jdx]
        self.delta_mag[idx] = delta_mag[jdx]
        self.gamma[idx] = gamma[jdx]
        self.vega[idx] = vega[jdx]
        self.moneyness[idx] = moneyness
        self.log_moneyness[idx] = log_moneyness
        self.standardised_moneyness[idx] = standardised_moneyness[jdx]
        self.OTM[idx] = OTM
        self.valid_price[idx]=valid_price
        self.underlying_px[idx]=underlying_px
    
    def update_with_instrument_object(self, idx, option_object, ivol, update_list):
            self.ivol[idx] = ivol
            for metric in update_list:    
                getattr(self, metric)[idx] = getattr(option_object, metric)
        

@dataclass(slots=True, frozen=True)
class BaseDomain:
    strike: np.ndarray
    expiry: np.ndarray
    x_metric: str
    y_metric: str
    
        
@dataclass(slots=True)
class Domain:
    base_domain: BaseDomain
    x: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y: np.ndarray = field(init=False, default_factory=lambda: np.array)
    x_vect: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y_vect: np.ndarray = field(init=False, default_factory=lambda: np.array)
    x_mat: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y_mat: np.ndarray = field(init=False, default_factory=lambda: np.array)
    z_mat: np.ndarray = field(init=False, default_factory=lambda: np.array)

    xy: np.ndarray = field(init=False, default_factory=lambda: np.array)
    idx_ij_mapping: Dict[int, tuple] = field(init=False, default_factory=dict)
    z_mat: float = field(init=False, default_factory=lambda: np.array)
    
    def __post_init__(self):
        self._create_domain(self.base_domain.strike.copy(), self.base_domain.expiry.copy())

    def _create_domain(self, x, y):
        self.x = x
        self.y = y
        self.xy = np.column_stack((self.x, self.y)) 

        self.x_vect = np.unique(self.base_domain.strike)
        self.y_vect = np.unique(self.base_domain.expiry)

        self.x_mat, self.y_mat = np.meshgrid(self.x_vect, self.y_vect, indexing="xy")
        self.z_mat = np.full(self.x_mat.shape, np.nan)
        
        self.idx_ij_mapping = {}
        
        for i in range(self.x_mat.shape[0]):
            for j in range(self.x_mat.shape[1]):
                match_idx = np.where((self.xy == [self.x_mat[i, j], self.y_mat[i, j]]).all(axis=1))[0]
                for idx in match_idx:
                    self.idx_ij_mapping[idx] = (i, j)

    def update_data(self, z, idx_map):
        for idx, ivol in zip(idx_map, z):
            if idx in self.idx_ij_mapping:
                i, j = self.idx_ij_mapping[idx]
                self.z_mat[i,j] = ivol

    def update(self, x=None, y=None, xy=None, z=None, x_metric=None, y_metric=None, z_metric=None):
        if not xy is None:
            self.xy = xy
            self.x = xy[:,0]
            self.y = xy[:,1]
        else:
            self.x = x
            self.y = y
            self.xy = np.column_stack((x,y))
        if not z is None:
            self.z = z
        if x_metric:
            self.x_metric = x_metric
        if y_metric:
            self.y_metric = y_metric
        self._calc_metrics()
    
    
@dataclass(slots=True)
class Scatter:  
    x: np.ndarray 
    y: np.ndarray 
    z: np.ndarray
    
    x_min: float = np.nan
    x_max: float = np.nan
    y_min: float = np.nan
    y_max: float = np.nan
    z_min: float = np.nan
    z_max: float = np.nan
    valid_values: bool = False
    
    x_min: float = field(init=False, default = np.nan)
    x_max: float = field(init=False, default = np.nan)
    y_min: float = field(init=False, default = np.nan)
    y_max: float = field(init=False, default = np.nan)
    z_min: float = field(init=False, default = np.nan)
    z_max: float = field(init=False, default = np.nan)
    
    valid_values: bool = field(init=False, default=False)
    
    def __post_init__(self):
        self.update_data(self.x, self.y, self.z)
        
    def get_limits(self,):
        return self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max
    
    def update_data(self, x, y, z):
        self.valid_values=False
        self.x, self.y, self.z = utils.filter_nans_on_z(x, y, z)
        if self.z.size > 0:
            self.valid_values=True
            self._calculate_metrics()

    def _calculate_metrics(self):
        self.x_min = np.amin(self.x)
        self.x_max = np.amax(self.x)
        self.y_min = np.amin(self.y)
        self.y_max = np.amax(self.y)
        self.z_min = np.amin(self.z)
        self.z_max = np.amax(self.z)

    def _calc_domain_metrics(self):
        self.x_min = np.amin(self.x)
        self.x_max = np.amax(self.x)
        self.y_min = np.amin(self.y)
        self.y_max = np.amax(self.y)

    def _calc_z_metrics(self):
        self.z_min = np.amin(self.z)
        self.z_max = np.amax(self.z)
                
        
@dataclass(slots=True)
class Surface:
    base_domain: BaseDomain
    scatter: Scatter
    interpolator: Any
    n_x: int
    n_y: int

    x: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y: np.ndarray = field(init=False, default_factory=lambda: np.array)
    z: np.ndarray = field(init=False, default_factory=lambda: np.array)
            
    x_min: float = field(init=False, default=np.nan)
    x_max: float = field(init=False, default=np.nan)
    y_min: float = field(init=False, default=np.nan)
    y_max: float = field(init=False, default=np.nan)
    
    z_min: float = field(init=False, default=np.nan)
    z_max: float = field(init=False, default=np.nan)
    
    valid_values: bool = field(init=False, default=False)
    last_interpolation_attempt_valid: bool = field(init=False, default=False)
    current_view_selection: str = field(init=False, default="Surface")
    
    def __post_init__(self):   
        self.x=self.scatter.x
        self.y=self.scatter.y
        self._create_interpolation_limits(self.scatter)
        if self.scatter.valid_values:
            self.interpolate_surface()
        else:
            self.z = np.nan * np.zeros((self.n_x, self.n_y))
        
    def get_limits(self,):
        return self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max
    
    def update_data_from_scatter_object(self, scatter_object):
        if (self.x_min != scatter_object.x_min
            or self.x_max != scatter_object.x_max
            or self.y_min != scatter_object.y_min
            or self.y_max != scatter_object.y_max):

            self.x_min = scatter_object.x_min 
            self.x_max = scatter_object.x_max
            self.y_min = scatter_object.y_min
            self.y_max = scatter_object.y_max

            self.x = np.linspace(self.x_min, self.x_max, self.n_x)
            self.y = np.linspace(self.y_min, self.y_max, self.n_y)

    def _create_interpolation_limits(self, scatter_object):
        self.x_min = self.base_domain.strike.min()
        self.x_max = self.base_domain.strike.max()
        self.y_min = self.base_domain.expiry.min()
        self.y_max = self.base_domain.expiry.max()

        """
        self.x_min = scatter_object.x_min
        self.x_max = scatter_object.x_max 
        self.y_min = scatter_object.y_min 
        self.y_max = scatter_object.y_max
        """
        self.x = np.linspace(self.x_min, self.x_max, self.n_x)
        self.y = np.linspace(self.y_min, self.y_max, self.n_y)

    def interpolate_surface(self):
        self.valid_values=False
        x, y, z = utils.filter_nans_on_z(self.scatter.x, self.scatter.y, self.scatter.z)
        if z.size < 5:
            warnings.warn(exceptions.InsufficientDataWarning(
                        f"Unable to interpolate: z has only {z.size} points (min 5 required)", 
                code=100
            ),
            stacklevel=2  # Point to the caller's line
        )
            self.last_interpolation_attempt_valid = False
            return
        try:
            self.interpolator.fit(x, y, z)                
        except Exception as e:
            print(f"The interpolator could not fit: {e}")
            self.last_interpolation_attempt_valid=False
            return
        try:
            self.z = self.interpolator.evaluate(self.x, self.y)
        except Exception as e:
            print(f"The interpolator could not evaluate: {e}")
            self.last_interpolation_attempt_valid=False
            return
        
        if not self.last_interpolation_attempt_valid:
            print("The last interpolation was valid")
        self.last_interpolation_attempt_valid=True
        self.z_min = np.nanmin(self.z)
        self.z_max = np.nanmax(self.z)
        self.valid_values=True

@dataclass
class Slice:
    x: np.ndarray
    y: np.ndarray
    xi: np.ndarray
    yi: np.ndarray
    valid_data: bool = field(init=False)
    interpolator: Any

    def __post_init__(self):
        self.x, self.y = utils.filter_nans_2D(self.x, self.y)
        if self.y.size > 1:
            self.interpolate(self.x, self.y)
            self.valid_data=True
        else:
            self.valid_data=False
        
    def interpolate(self, x, y):
        self.x, self.y = utils.filter_nans_2D(self.x, self.y)
        if self.y.size > 2:
            self.interpolator.fit(x, y)
            self.yi = self.interpolator.evaluate(self.xi)
            self.valid_data=True
        else:
            self.valid_data = False


@dataclass
class Smile:
    base_domain: BaseDomain
    raw: RawOptions
    interpolator: Any
    n_x: int
    slice: Dict[Slice]
    
    x: np.ndarray = field(init=False)
    z_dict: Dict[np.ndarray] = field(init=False)
    x_min: float = field(init=False)
    x_max: float = field(init=False)
    expiry_vect: np.ndarray = field(init=False)
    
    def __post_init__(self):
        self.expiry_vect = np.unique(self.base_domain.expiry)
        x = self.base_domain.strike
        y = self.base_domain.expiry
        
        self.x_min = x.min()
        self.x_max = x.max()
        self.x = np.linspace(self.x_min, self.x_max, self.n_x)
        self.z_dict = {exp : [] for exp in self.expiry_vect}
        self.slice = {}
        for expiry in self.expiry_vect:
             slice = Slice(x, y, self.x, np.full(self.x.shape, np.nan))
             self.slice[expiry] = slice
             
    def update_data(self, x, y, z):
        if z.size > 2:
            pass
        
     
    def interpolate(self):
        x = self.base_domain.strike
        y = self.base_domain.expiry
        z = self.raw.ivol
        x = x[self.raw.OTM]
        y = y[self.raw.OTM]
        z = z[self.raw.OTM]
        x, y, z = utils.filter_nans_on_z(x, y, z)
        for expiry in self.expiry_vect:
            mask = y == expiry
            x_filtered = x[mask]
            z_filtered = z[mask]
            slice = self.slice[expiry]
            slice.update_data(x_filtered, z_filtered)   
            if z_filtered.size > 3:
                self.interpolator.fit(x_filtered, z_filtered)
                z_new = self.interpolator.evaluate(self.x)
                self.z_dict[expiry] = z_new
                self.valid_data_dict[expiry] = True
            else:
                self.valid_data_dict[expiry] = False


@dataclass
class DataContainer:
    px_type: str
    domain: Optional["Domain"] = None
    raw: Optional["RawOptions"] = None
    scatter: Optional["Scatter"] = None
    surface: Optional["Surface"] = None
    smile: Optional["Smile"] = None
        
    def __post_init__(self):
        if not self.surface is None and not self.scatter is None:
            self._calculate_data_limits()
    
    def update_dataclasses(self, x, y, z, idx_map):
        if not self.domain is None:
            self.domain.update_data(z, idx_map)
        if not self.scatter is None:
            self.scatter.update_data(x, y, z)
        if not self.surface is None:
            self.surface.update_data_from_scatter_object(self.scatter)

    def update_dataclasss_transform(self, transform_engine):
        x, y, z, idx_map = transform_engine.transform_data(self.raw)
        self.update_dataclasses(x, y, z, idx_map)
        
    def data_container_cleanup(self):
        if not self.surface is None:
            self.surface.interpolate_surface()
        if not self.smile is None:
            self.smile.interpolate()
        self._calculate_data_limits()
    
    def _calculate_data_limits(self):
        self.x_min = np.minimum(self.scatter.x_min, self.surface.x_min)
        self.x_max = np.maximum(self.scatter.x_max, self.surface.x_max) 
        self.y_min = np.minimum(self.scatter.y_min, self.surface.y_min)
        self.y_max = np.maximum(self.scatter.y_max, self.surface.y_max) 
        
        if self.scatter.valid_values and self.surface.valid_values:
            self.z_min = np.minimum(self.scatter.z_min, self.surface.z_min)
            self.z_max = np.maximum(self.scatter.z_max, self.surface.z_max) 
        elif self.scatter.valid_values:
            self.z_min = self.scatter.z_min
            self.z_max = self.scatter.z_max
        elif self.surface.valid_values:
            self.z_min=self.surface.z_min
            self.z_max=self.surface.z_max
        

@dataclass
class DataFeatureManager:
    price_types: list

    limit_dict: dict = field(init=False)
    valid_values_dict: dict = field(init=False)
    valid_values_any: bool = False
    
    all_axis: ClassVar[list] = ["x","y","z"]
    plot_types: ClassVar[list] = ["scatter", "surface"]
    _reset_dict: ClassVar[dict] = {"min" : np.nan, "max" : np.nan}
    
    def __post_init__(self):
        self.limit_dict = {}
        self.valid_values_dict = {}
        for px_type in self.price_types:
            plot_dict_lim = {}
            plot_dict_bool = {}
            for plot_type in self.plot_types:
                axis_dict_lim = {}
                for axis in self.all_axis:
                    axis_dict_lim[axis] = copy.deepcopy(self._reset_dict)
                plot_dict_lim[plot_type] = axis_dict_lim
                plot_dict_bool[plot_type] = False
                
            self.limit_dict[px_type] = plot_dict_lim
            self.valid_values_dict[px_type] = plot_dict_bool

    def remove_data(self, px_type):
        for axis in self.all_axis:
            self.limit_dict[px_type]["surface"][axis] = copy.deepcopy(self._reset_dict)
            self.limit_dict[px_type]["scatter"][axis] = copy.deepcopy(self._reset_dict)
            
        self.valid_values_dict[px_type]["surface"]=False
        self.valid_values_dict[px_type]["scatter"]=False

    def find_any_valid_values(self, data_container_manager_object: DataContainerManager.object):
        bool_arr = []
        z_vals = []
        for px_type, data_container in data_container_manager_object.items():
            if data_container.surface.valid_values:
                for axis in self.all_axis:
                    axis_min = getattr(data_container.surface, f"{axis}_min")
                    axis_max = getattr(data_container.surface, f"{axis}_max")
                    self.limit_dict[px_type]["surface"][axis]["min"] = axis_min
                    self.limit_dict[px_type]["surface"][axis]["max"] = axis_max
                    bool_arr.append(True)
                    self.valid_values_dict[px_type]["surface"] = True
                z_vals.append(self.limit_dict[px_type]["surface"]["z"]["min"])
                z_vals.append(self.limit_dict[px_type]["surface"]["z"]["max"])
            else:
                for axis in self.all_axis:
                    self.limit_dict[px_type]["surface"][axis]["min"] = np.nan
                    self.limit_dict[px_type]["surface"][axis]["max"] = np.nan
                    self.valid_values_dict[px_type]["surface"] = False

            if data_container.scatter.valid_values:
                for axis in self.all_axis:
                    axis_min = getattr(data_container.scatter, f"{axis}_min")
                    axis_max = getattr(data_container.scatter, f"{axis}_max")
                    self.limit_dict[px_type]["scatter"][axis]["min"] = axis_min
                    self.limit_dict[px_type]["scatter"][axis]["max"] = axis_max
                    bool_arr.append(True)
                    self.valid_values_dict[px_type]["scatter"] = True
                z_vals.append(self.limit_dict[px_type]["scatter"]["z"]["min"])
                z_vals.append(self.limit_dict[px_type]["scatter"]["z"]["max"])
            else:
                self.limit_dict[px_type]["scatter"][axis]["min"] = np.nan
                self.limit_dict[px_type]["scatter"][axis]["max"] = np.nan
                self.valid_values_dict[px_type]["scatter"] = False
        
        self.valid_values_any = any(bool_arr)
        if self.valid_values_any:
            return True, z_vals
        else:
            return False, z_vals
          
                                                      
@dataclass
class DataContainerManager:
    axis_transform_engine: "MetricEngine" 
    price_types: InitVar[list]
    data_container: InitVar[DataContainer] = None
    
    containers: dict = field(init=False, default_factory=dict)
    displayed_price_types: List[str] = field(init=False, default_factory=list)  
    
    x_min: float = field(init=False, default=np.nan)
    x_max: float = field(init=False, default=np.nan)
    y_min: float = field(init=False, default=np.nan)
    y_max: float = field(init=False, default=np.nan)
    z_min: float = field(init=False, default=np.nan)
    z_max: float = field(init=False, default=np.nan)
    
    features: DataFeatureManager = None
    
    def __post_init__(self, price_types, data_container: DataContainer):
        self.features = DataFeatureManager(price_types)
        if not data_container is None:
            self.add_displayed_price_type(data_container)

    def add_displayed_price_type(self, px_type):
        self.displayed_price_types.append(px_type)
        self.calculate_data_limits()
    
    def add_data_container(self, data_container):
        self.containers[data_container.px_type] = data_container
        self.displayed_price_types = [data_container.px_type]
        self.calculate_data_limits()
        
    def remove_displayed_price_type(self, px_type):
        self.displayed_price_types.remove(px_type)
        self.features.remove_data(px_type)
        self.calculate_data_limits()

    def update_data_containers_instrument_object(self, instrument_object):
        for px_type, data_container in self.containers.items():
            data_container.raw.update_all_metrics_name(instrument_object.ticker,
                                                       *instrument_object.get_all_metrics_price_type(px_type))
    
    def update_data_containers(self):
        for px_type, data_container in self.containers.items():
            data_container.update_dataclasss_transform(self.axis_transform_engine)
            data_container.data_container_cleanup()
    
    def update_cleanup(self):
        for px_type, data_container in self.containers.items():
            data_container.update_dataclasss_transform(self.axis_transform_engine)
    
    def update_price_cleanup(self):
        self.calculate_data_limits()

    def process_update(self):
        if len(self.containers) > 0:
            self.calculate_data_limits()
            
    def calculate_data_limits(self):
        if len(self.displayed_price_types) > 0:
            self.x_min = np.min([data_container.x_min for data_container in self.containers.values() if data_container.px_type in self.displayed_price_types])
            self.x_max = np.max([data_container.x_max for data_container in self.containers.values() if data_container.px_type in self.displayed_price_types])
            self.y_min = np.min([data_container.y_min for data_container in self.containers.values() if data_container.px_type in self.displayed_price_types])
            self.y_max = np.max([data_container.y_max for data_container in self.containers.values() if data_container.px_type in self.displayed_price_types])
            valid_values_any, z_vals = self.features.find_any_valid_values(self.containers)
            if valid_values_any:
                self.z_min, self.z_max = np.amin(z_vals), np.amax(z_vals)
            else:
                self.z_min, self.z_max = np.nan, np.nan
        else:
            self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max = [np.nan] * 6
            self.valid_values_any=False
            
    def get_limits(self):
        return self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max
    
    def generate_mask_from_domain_lims(self, x, y):
        return (self.x_min < x) & (x < self.x_max) & (self.y_min < y) & (y < self.y_max)



# Copyright (C) 2013 Riverbank Computing Limited.
# Copyright (C) 2022 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR BSD-3-Clause
from __future__ import annotations

"""PySide6 port of the widgets/layouts/flowlayout example from Qt v6.x"""

import sys
from PySide6.QtCore import Qt, QMargins, QPoint, QRect, QSize
from PySide6.QtWidgets import QApplication, QLayout, QPushButton, QSizePolicy, QWidget
from PySide6 import QtCore, QtGui, QtWidgets

class Window(QWidget):
    def __init__(self):
        super().__init__()

        flow_layout = FlowLayout(self)
        flow_layout.addWidget(QPushButton("Short"))
        flow_layout.addWidget(QPushButton("Longer"))
        flow_layout.addWidget(QPushButton("Different text"))
        flow_layout.addWidget(QPushButton("More text"))
        flow_layout.addWidget(QPushButton("Even longer button text"))

        self.setWindowTitle("Flow Layout")



class FlowLayout(QtWidgets.QLayout):
    def __init__(self, parent=None, margin=0, spacing=-1):
        super().__init__(parent)
        
        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)
            
        self.setSpacing(spacing)
        self.itemList = []
        
    def addItem(self, item):
        self.itemList.append(item)
        
    def count(self):
        return len(self.itemList)
        
    def itemAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList[index]
        return None
        
    def takeAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList.pop(index)
        return None
        
    def expandingDirections(self):
        return QtCore.Qt.Orientations(QtCore.Qt.Orientation(0))
        
    def hasHeightForWidth(self):
        return True
        
    def heightForWidth(self, width):
        return self.doLayout(QtCore.QRect(0, 0, width, 0), True)
        
    def setGeometry(self, rect):
        super().setGeometry(rect)
        self.doLayout(rect, False)
        
    def sizeHint(self):
        return self.minimumSize()
        
    def minimumSize(self):
        size = QtCore.QSize()
        
        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())
            
        margins = self.contentsMargins()
        size += QtCore.QSize(margins.left() + margins.right(), margins.top() + margins.bottom())
        return size
        
    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0
        
        for item in self.itemList:
            wid = item.widget()
            spaceX = self.spacing()
            spaceY = self.spacing()
            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0
                
            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))
                
            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())
            
        return y + lineHeight - rect.y()

if __name__ == "__main__":
    app = QApplication(sys.argv)
    main_win = Window()
    main_win.show()
    sys.exit(app.exec())





import sys
import weakref
import numpy as np
from OpenGL.GL import * 
import pyqtgraph as pg
from pyqtgraph import opengl
import warnings
from PySide6 import QtWidgets, QtCore, QtGui
from pyvol_terminal.gl_3D_graphing import GLGraphicsItem
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem as pyqtGLGraphicsItem
from pyvol_terminal.gl_3D_graphing.widgets.GLScatterPlotItem import GLScatterPlotItem 
import PySide6

class CustomGraphicsItem(GLGraphicsItem.GLGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)


class AnotherGraphicsItem(CustomGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)
    
    def boundingRect(self):
        return self.childrenBoundingRect()


class ViewBox(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__()
        self.__view=None
        self.__initialized=False
        self.other_graphics_item = AnotherGraphicsItem(parent=self)
        self.setGroup(QtWidgets.QGraphicsItemGroup())
    
    def boundingRect(self):
        return self.childrenBoundingRect()
    
    def _setView(self, v):
        self.__view = v

    def initialize(self):
        self.initializeGL()
        self.__initialized = True

    def isInitialized(self):
        return self.__initialized
    
    def initializeGL(self):
        pass

class CustomGLViewWidget(opengl.GLViewWidget):  
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def setViewBox(self, vb):
        self.vb=vb

class Window(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.vb = ViewBox()
        self.gl_view = CustomGLViewWidget()
        self.setCentralWidget(self.gl_view)
        self.show()
        print(self.gl_view.isValid())
        
        self.view_box = ViewBox()
        self.gl_view.addItem(self.view_box)
        print(self.gl_view.items)

        
        x_range, y_range = [-5, 5], [-5, 5]
        self.scatter, self.surface = self.create_surface(x_range, y_range)
        self.gl_view.addItem(self.scatter)
        self.gl_view.addItem(self.surface)
        self.update()
        
        self.view_box.setParentItem(self.scatter)
        self.view_box.setParentItem(self.surface)

        
        
    def create_surface(self, x_range, y_range):
        X = np.arange(*x_range, 0.5)
        Y = np.arange(*y_range, 0.5)
        X_mat, Y_mat = np.meshgrid(X, Y)
        
        R_shifted = np.sqrt((X_mat)**2 + (Y_mat)**2)
        Z_mat = 2 + np.sin(R_shifted)                
        X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
        
        pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
        pos_surface = [X, Y, 3*Z_mat]

        
        scatter = GLScatterPlotItem(pos=pos_scatter, color=(1, 0, 0, 1))
        surface = opengl.GLSurfacePlotItem(*pos_surface,
                                    glOptions='opaque',
                                    color=(0.5, 0.5, 1, 1))
        
        return scatter, surface



def main():
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()
    



from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from pyqtgraph.opengl import GLLinePlotItem
import sys


from pyqtgraph import Transform3D

import numpy as np
from pyqtgraph.opengl import GLViewWidget

from PySide6 import QtCore, QtGui, QtWidgets
import numpy as np
from pyqtgraph.opengl import GLLinePlotItem, GLGraphicsItem
from pyqtgraph import Transform3D

class TransformedLineGroup(GLGraphicsItem):
    def __init__(self):
        super().__init__()
        self.lines = []  # Stores (vertices, color, transform)
        self.line_item = GLLinePlotItem()  # Single renderable item

    def addLine(self, vertices, color, transform=None):
        """Add a line with optional transform."""
        if transform is None:
            transform = Transform3D()  # Identity if no transform given
        self.lines.append((vertices, color, transform))
        self.updateLineData()  # Rebuild merged data

    def updateLineData(self):
        """Combine all lines into a single vertex array."""
        if not self.lines:
            return

        # Apply transforms and concatenate vertices
        all_vertices = []
        all_colors = []
        for vertices, color, transform in self.lines:
            # Apply transform to each vertex
            transformed_vertices = np.array([transform.map(v) for v in vertices])
            all_vertices.append(transformed_vertices)
            # Repeat color for each vertex
            all_colors.append(np.tile(color, (len(vertices), 1)))

        # Concatenate into single array
        merged_vertices = np.vstack(all_vertices)
        merged_colors = np.vstack(all_colors)

        # Update the line plot item
        self.line_item.setData(pos=merged_vertices, color=merged_colors)

    def paint(self):
        """Render all lines in one call."""
        self.line_item.paint()

class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, app=None, **kwargs):
        super().__init__(*args, **kwargs)   

def main():
    
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    win.timer.start(500)
    sys.exit(app.exec())

if __name__ == "__main__":
    main()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Cfullable
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.widgets.view_box import ViewBox, AxisNormaliser

from PySide6 import QtWidgets, QtCore
import sys
from pyvol_terminal.gl_3D_graphing.widgets import GL3DViewWidget
from pyqtgraph import opengl
import numpy as np
import uuid 
from dataclasses import dataclass, field


@dataclass(slots=True)
class PlotDataset:
    full: Dict[int, np.ndarray]
    _subset: Dict[int, np.ndarray] = field(init=False, default=None)
    limits_full: Dict[int, Tuple[float, float]] = field(init=False, default_factory=lambda: {ax : (None, None) for ax in range(3)})
    limits_subset: Dict[int, Tuple[float, float]] = field(init=False, default=None)
    _subset_mask: Dict[int, np.ndarray] = field(init=False, default=None)
    _limits_used_for_subset: Dict[int, List[float, float]] = field(init=False, default=None)
    _subset_on: bool = False
    

    def __post_init__(self):
        self.resetSubset()
        
    def setFull(self, full):
        self.full=full
        self._computerLimitsFull()

    def setSubset(self, subset):
        self._subset=subset
        self._computerLimitsSubset()

    def setFullForAxis(self, ax, arr):
        self.full[ax]=arr
        self._computerLimitsFullForAxis(ax)
        
    def updateDatasetAxis(self, ax, arr):
        self.setFullForAxis(ax, arr)
        if self._subset_on:
            self.computeSubsetFromLimits(ax, self._subset_mask[ax])
        self.setSubsetForAxis(ax, arr)        
        
    def setSubsetForAxis(self, ax, arr):
        self.full[ax]=arr
        
    def _axisSubsetCleaup(self, ax):
        self._computerLimitsSubsetForAxis(ax)
        subset_dims_equal_full = [self._subset_mask[ax].sum() == self.full[ax].size for ax in range(3)]
        if all((any(subset_dims_equal_full),
                any([not a is None for a in self._limits_used_for_subset[2]]),
                any((self.limits_full[2][0] != self.limits_subset[2][0],
                     self.limits_full[2][1] != self.limits_subset[2][1]))
                )):
            self._subset_on=True
        else:
            self._subset_on=False            
        
    def subset(self, ax=None):
        if ax is None:
            return self._subset[0], self._subset[1], self._subset[2]
        else:
            return self._subset[ax]
        
    def computeSubsetFromLimits(self, ax, limits):
        self._limits_used_for_subset[ax]=limits
        z_prev_lims = self._limits_used_for_subset[2]
        if ax < 2:
            mask_ax = (self.full[ax] >= limits[0]) & (self.full[ax] <= limits[1])
            new_subset = self.full[ax][mask_ax]
            self._subset_mask[ax]=mask_ax
            mask_other = self._subset_mask[1-ax]
            if ax == 0:
                new_subset_z = self.full[2][mask_ax,:][:,mask_other]
            elif ax == 1:
                new_subset_z = self.full[2][mask_other,:][:,mask_ax]
            
            z_clipped = np.clip(new_subset_z, z_prev_lims[0], z_prev_lims[1])
            self._subset[2] = z_clipped
            self._axisSubsetCleaup(2)
        else:
            new_subset_z = self.full[2][self._subset_mask[0],:][:,self._subset_mask[1]]
            new_subset = np.clip(new_subset_z, z_prev_lims[0], z_prev_lims[1])
        self._subset[ax]=new_subset
        self._axisSubsetCleaup(ax)

    def fullRange():
        return self._ran
        
    def resetSubset(self):
        self._computerLimitsFull()
        self._subset = {ax : arr.copy() for ax, arr in self.full.items()}
        self.limits_subset = {ax : (limit[0], limit[1]) for ax, limit in self.limits_full.items()}
        self._subset_mask = {ax : np.ones(arr.shape, dtype=bool) for ax, arr in self.full.items()}
        self._subset_on=False
        self._limits_used_for_subset = {ax : (None, None) for ax in range(3)}
    
    def _computerLimitsFull(self):
        for ax in self.full:
            self._computerLimitsFullForAxis(ax)

    def _computerLimitsSubset(self):
        for ax, arr in self._subset.items():
            self.limits_subset[ax] = np.nanmin(arr), np.nanmax(arr)
            
    def _computerLimitsFullForAxis(self, axis):
        self.limits_full[axis] = np.nanmin(self.full[axis]), np.nanmax(self.full[axis])

    def _computerLimitsSubsetForAxis(self, axis):
        self.limits_subset[axis] = np.nanmin(self._subset[axis]), np.nanmax(self._subset[axis])
        



class CustomSurface(opengl.GLSurfacePlotItem):
    def __init__(self, *args, **kwargs):
        self._view_box: ViewBox=None
        self.__parent = None
        self.__view=None
        self._id = uuid.uuid4()
        
        full = {idx : arr for idx, arr in enumerate(args[:3])}
        
        self.dataset = PlotDataset(full)
        
        super().__init__(*args, **kwargs)
        
        self._ax_idx_str_map = {i : s for i, s in enumerate("xyz")}
        
            
    def dataInView(self,
                   view_box: ViewBox
                   ):
        setdataKwargs = {}
        zvalues_in_view=None
        for ax in range(3):
            viewRange = view_box._viewRange[ax]
            if ax != 2:
                values = self.dataset.full[ax]
                mask = (values >= viewRange[0]) & (values <= viewRange[1])
                values_in_view = values[mask]
                setdataKwargs[self._ax_idx_str_map[ax]] = values_in_view
                self.dataset.setSubsetForAxis(ax, values_in_view)
                
                if ax==0:
                    values_z = self.dataset.full[2].copy()
                    zvalues_in_view = values_z[mask, :]
                else:
                    zvalues_in_view = zvalues_in_view[:, mask]
                setdataKwargs["z"] = zvalues_in_view
                
            else:
                zvalues_in_view = np.clip(zvalues_in_view, viewRange[0], viewRange[1])
                setdataKwargs["z"] = zvalues_in_view
                self.dataset.setSubsetForAxis(2, zvalues_in_view)
        return setdataKwargs
    
    def updateXView(self, view_box: ViewBox):
        xRange = view_box._viewRange[0]
        self.dataset.computeSubsetFromLimits(0, xRange)
        x_norm, y_norm, z_norm = view_box.normaliseToView(*self.dataset.subset())
        super().setData(x=x_norm, y=y_norm, z=z_norm)        
        
    def updateYView(self, view_box: ViewBox):
        yRange = view_box._viewRange[1]
        self.dataset.computeSubsetFromLimits(1, yRange)
        x_norm, y_norm, z_norm = view_box.normaliseToView(*self.dataset.subset())
        super().setData(x=x_norm, y=y_norm, z=z_norm)
    
    def updateZView(self, view_box: ViewBox):
        zRange = view_box._viewRange[2]
        self.dataset.computeSubsetFromLimits(2, zRange)
        x_norm, y_norm, z_norm = view_box.normaliseToView(*self.dataset.subset())
        super().setData(x=x_norm, y=y_norm, z=z_norm)

    def updateInView(self,
                     ax: int,
                     view_box: ViewBox
                     ):
        setdataKwargs = self.dataInView(view_box)
        super().setData(**setdataKwargs)


    def linktoView(self,
                   view_box: ViewBox):
        self._view_box=view_box
    
    def filterinView(self, x=None, y=None, z=None):
        if self._view_box is None:
            return {"x" : x, "y" : y, "z" : z}
        setDataKwargs = self.dataInView(self._view_box)
        return setDataKwargs
    
    def _update_dataset(self, x=None, y=None, z=None):
        if not x is None:
            self.dataset.updateDatasetAxis(0, x)
        if not y is None:
            self.dataset.updateDatasetAxis(1, y)
        if not z is None:
            self.dataset.updateDatasetAxis(2, z)

    def setData(self, x=None, y=None, z=None, colors=None):
        self._update_dataset(x, y, z)
        if not self._view_box is None:
            x_norm, y_norm, z_norm = self._view_box.normaliseToView(*self.dataset.subset())
        else:
            x_norm, y_norm, z_norm = self.dataset.subset()
        super().setData(x=x_norm, y=y_norm, z=z_norm)
        
    def setDataFromView(self, view_box: ViewBox):
        for i, viewRange in enumerate(view_box.viewRange()):
            self.dataset.computeSubsetFromLimits(i, viewRange)
        x_norm, y_norm, z_norm = view_box.normaliseToView(*self.dataset.subset())
        super().setData(x=x_norm, y=y_norm, z=z_norm)
        
    def id(self):
        return self._id


class CustomScatter(opengl.GLScatterPlotItem):
    def __init__(self, *args, **kwargs):
        self._view_box: ViewBox=None
        self.__parent = None
        self.__view=None
        self._id = uuid.uuid4()

        super().__init__(*args, **kwargs)

    def setData(self, **kwds):
        super().setData(**kwds)
        if not self._view_box is None:
             self.setTransform(self._view_box.transform_matrix)

    def id(self):
        return self._id


class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        central = QtWidgets.QWidget()
        self.setCentralWidget(central)
        self.central_layout = QtWidgets.QVBoxLayout()
        central.setLayout(self.central_layout)
        self.gl_view = GL3DViewWidget.GL3DViewWidget(padding=[0.]*3)
        X = np.arange(-5, 5, 0.5)
        Y = np.arange(-5, 5, 0.5)
        X_mat, Y_mat = np.meshgrid(X, Y)

        Z_mat = 3 * np.sin(X_mat) * np.cos(Y_mat)
        X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
        
        pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
        pos_surface = [X, Y, Z_mat]

        
        
      #  scatter = CustomScatter(pos=pos_scatter, color=(1, 0, 0, 1))
        surface = CustomSurface(*pos_surface,
                                glOptions='opaque',
                                color=(0.5, 0.5, 1, 1)
                                )
       # self.gl_view.addItem(scatter)
        self.gl_view.addItem(surface)

       # self.scatter=scatter
        self.surface=surface
        
        #scatter = CustomScatter(pos=pos)
        
        
        self.central_layout.addWidget(self.gl_view)     
    #    z_new = surface._z * 2
     #   surface.setData(z=z_new)
        #self.gl_view.addItem(scatter, ignoreBounds=False) 

        for i in range(1, 11):
            x = opengl.GLTextItem(pos = (i/10, 0, 0), text=f"{round(i/10, 1)}", color="white")
            y = opengl.GLTextItem(pos = (0, i/10, 0), text=f"{round(i/10, 1)}", color="yellow")
            z = opengl.GLTextItem(pos = (i/10, 0, i/10), text=f"{round(i/10, 1)}", color="cyan")
            self.gl_view.addItem(x, ignoreBounds=True)
            self.gl_view.addItem(y, ignoreBounds=True)
            self.gl_view.addItem(z, ignoreBounds=True)
            
        self.showMaximized()
        
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update)
        #self.timer.start(500)
        
    def update(self):
        i,j=np.unravel_index(np.argmax(self.surface._z), self.surface._z.shape)
        z_new= self.surface._z.copy()
        z_new[i,j] = z_new[i,j]+0.5
        self.surface.setData(z=z_new)
        
        pos = self.scatter.pos.copy()
        idx = np.argmax(pos[:,0])
        pos[idx,0] = pos[idx,0]+0.1
        self.scatter.setData(pos=pos)
        #self.gl_view.vb.normalize_scene()



def main():
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()
    

    



from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
import pyqtgraph.opengl as gl
import pyqtgraph as pg
import OpenGL.GL as ogl
import numpy as np
import sys

class CustomTextItem(gl.GLGraphicsItem.GLGraphicsItem):
    def __init__(self, X, Y, Z, text):
        self.color = QtCore.Qt.GlobalColor.white
        gl.GLGraphicsItem.GLGraphicsItem.__init__(self)
        self.text = text
        self.X = X
        self.Y = Y
        self.Z = Z
        self.color = QtCore.Qt.GlobalColor.white
        self.font = QtGui.QFont('Helvetica', 16)

        self._setPos()
    
    def _setPos(self):
        self.pos=(self.X, self.Y, self.Z)

    def setGLViewWidget(self, GLViewWidget):
        self.GLViewWidget = GLViewWidget

    def setText(self, text):
        self.text = text
        self.update()

    def setX(self, X):
        self.X = X
        self._setPos()
        self.update()

    def setY(self, Y):
        self.Y = Y
        self._setPos()
        self.update()

    def setZ(self, Z):
        self.Z = Z
        self._setPos()
        self.update()

    def paint(self):
      #  self.GLViewWidget.qglColor(QtCore.Qt.black)
        if len(self.text) < 1:
            return
        self.setupGLState()

        project = self.compute_projection()
        vec3 = QtGui.QVector3D(*self.pos)
        text_pos = project.map(vec3).toPointF()

        painter = QtGui.QPainter(self.view())
        painter.setPen(self.color)
        painter.setFont(self.font)
        painter.setRenderHints(QtGui.QPainter.RenderHint.Antialiasing | QtGui.QPainter.RenderHint.TextAntialiasing)
        painter.drawText(text_pos, self.text)
        painter.end()
        
    def compute_projection(self):
        # note that QRectF.bottom() != QRect.bottom()
        rect = QtCore.QRectF(self.view().rect())
        ndc_to_viewport = QtGui.QMatrix4x4()
        ndc_to_viewport.viewport(rect.left(), rect.bottom(), rect.width(), -rect.height())
        return ndc_to_viewport * self.mvpMatrix()

class Custom3DAxis(gl.GLAxisItem):
    """Class defined to extend 'gl.GLAxisItem'."""
    def __init__(self, parent, color=(0,0,0,.6)):
        self.antialias=False
        gl.GLAxisItem.__init__(self)
        self.parent = parent
        self.color = color
        

    def add_labels(self):
        """Adds axes labels."""
        x,y,z = self.size()
        #X label
        self.xLabel = CustomTextItem(X=x/2, Y=-y/20, Z=-z/20, text="X")
        self.xLabel.setGLViewWidget(self.parent)
        self.parent.addItem(self.xLabel)
        #Y label
        self.yLabel = CustomTextItem(X=-x/20, Y=y/2, Z=-z/20, text="Y")
        self.yLabel.setGLViewWidget(self.parent)
        self.parent.addItem(self.yLabel)
        #Z label
        self.zLabel = CustomTextItem(X=-x/20, Y=-y/20, Z=z/2, text="Z")
        self.zLabel.setGLViewWidget(self.parent)
        self.parent.addItem(self.zLabel)

    def add_tick_values(self, xticks=[], yticks=[], zticks=[]):
        """Adds ticks values."""
        x,y,z = self.size()
        xtpos = np.linspace(0, x, len(xticks))
        ytpos = np.linspace(0, y, len(yticks))
        ztpos = np.linspace(0, z, len(zticks))
        #X label
        for i, xt in enumerate(xticks):
            val = CustomTextItem(X=xtpos[i], Y=-y/20, Z=-z/20, text=str(xt))
            val.setGLViewWidget(self.parent)
            self.parent.addItem(val)
        #Y label
        for i, yt in enumerate(yticks):
            val = CustomTextItem(X=-x/20, Y=ytpos[i], Z=-z/20, text=str(yt))
            val.setGLViewWidget(self.parent)
            self.parent.addItem(val)
        #Z label
        for i, zt in enumerate(zticks):
            val = CustomTextItem(X=-x/20, Y=-y/20, Z=ztpos[i], text=str(zt))
            val.setGLViewWidget(self.parent)
            self.parent.addItem(val)

    def paint(self):
        self.setupGLState()
        if self.antialias:
            ogl.glEnable(ogl.GL_LINE_SMOOTH)
            ogl.glHint(ogl.GL_LINE_SMOOTH_HINT, ogl.GL_NICEST)
        ogl.glBegin(ogl.GL_LINES)

        x,y,z = self.size()
        #Draw Z
        ogl.glColor4f(*self.color)
        ogl.glVertex3f(0, 0, 0)
        ogl.glVertex3f(0, 0, z)
        #Draw Y
        ogl.glColor4f(*self.color)
        ogl.glVertex3f(0, 0, 0)
        ogl.glVertex3f(0, y, 0)
        #Draw X
        ogl.glColor4f(*self.color)
        ogl.glVertex3f(0, 0, 0)
        ogl.glVertex3f(x, 0, 0)
        ogl.glEnd()

def main():
    app = QtWidgets.QApplication(sys.argv)
    fig1 = gl.GLViewWidget()
    n = 51
    y = np.linspace(-10,10,n)
    x = np.linspace(-10,10,100)
    for i in range(n):
        yi = np.array([y[i]]*100)
        d = (x**2 + yi**2)**0.5
        z = 10 * np.cos(d) / (d+1)
        pts = np.vstack([x,yi,z]).transpose()
        plt = gl.GLLinePlotItem(pos=pts, color=pg.glColor((i,n*1.3)), width=(i+1)/10., antialias=True)
        fig1.addItem(plt)


    axis = Custom3DAxis(fig1, color=(0.2,0.2,0.2,.6))
    axis.setSize(x=12, y=12, z=12)
    # Add axes labels
    axis.add_labels()
    # Add axes tick values
    axis.add_tick_values(xticks=[0,4,8,12], yticks=[0,6,12], zticks=[0,3,6,9,12])
    fig1.addItem(axis)
    fig1.opts['distance'] = 40

    fig1.show()
    sys.exit(app.exec())

if __name__ == '__main__':
    main()


from pyvol_terminal.engines.option_engines import base
from pyvol_terminal.engines.yield_engines import implied_engines


def update_config(config):
    _check_data_config(config)
    
    if not "starting_price_type" in config: 
        config["starting_price_type"]="mid"
    
    if "df_spot" in config["data_config"]:
        config["spot_flag"]=True
    else:
        config["spot_flag"]=False
    
    if not "colour_styles_config" in config:
        config["colour_styles_config"] = {"scatter" : {"bid" : (1,0,0,1), "ask" : (0, 0, 1, 1), "mid" : (1, 1, 1, 1)},
                                          "surface" : {"bid" : "inferno", "ask" : "inferno", "mid" : "inferno"},
                                         }
    _check_option_config(config)
    _check_yield_configs(config)


def _check_data_config(config):
    data_config = config["data_config"]
    
    if not "df_options" in data_config and not "df_futures" in data_config:
        raise
    
def _check_yield_configs(config):
    if not "interest_rate_config" in config:
        config["interest_rate_config"] = {"engine" : implied_engines.DummyYieldClass(),
                                          "use_ws_response" : False}
    else:
        if not "engine" in config["interest_rate_config"]:
            config["interest_rate_config"] = {"engine" : implied_engines.DummyYieldClass(),
                                               "use_ws_response" : False}
            
    if not "dividend_rate_config" in config:
        config["dividend_rate_config"] = {"engine" : implied_engines.DummyYieldClass(),
                                          "use_ws_response" : False}
    else:
        if not "engine" in config["dividend_rate_config"]:
            config["dividend_rate_config"] = {"engine" : implied_engines.DummyYieldClass(),
                                              "use_ws_response" : False}

    
def _check_option_config(config):
    pass





from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from uuid import uuid4
    from ...interfaces.custom_widgets import CustomPlotDataItem
    from ..graphics_items.GL3DAxisItem import GL3DAxisItem
    from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GL3DViewBox
    
import numpy as np
from pyqtgraph import opengl
from PySide6 import QtGui, QtCore, QtWidgets
from PySide6.QtOpenGLWidgets import QOpenGLWidget
from ..graphics_items.GL3DPlotDataItemMixin import BaseGL3DPlotDataItemMixin
from ..graphics_items import GL3DViewBox
import math
from OpenGL import GL
from OpenGL import GLU
import weakref
from pyqtgraph.Qt import isQObjectAlive
from OpenGL.GL import glEnable, GL_DEPTH_TEST
from pyqtgraph import ButtonItem, icons, Vector
from pprint import pprint
import warnings
import time

class GLButtonitem(QtWidgets.QPushButton):
    def __init__(self, plotitem_widget: QOpenGLWidget, *args, **kwargs):
        icon_pixmap = icons.getGraphPixmap('auto')
        super().__init__(plotitem_widget)
        self.setIcon(QtGui.QIcon(icon_pixmap))
        self.setIconSize(QtCore.QSize(30, 30))
        self.setFixedSize(30, 30)
        self.setStyleSheet("""
                            QPushButton {
                                background: transparent;
                                border: none;
                            }
                            QPushButton:hover {
                                background: rgba(200, 200, 200, 100);
                                border-radius: 5px;
                            }
                            """)
                        
        self.updatePosition()
        plotitem_widget.resized.connect(self.updatePosition)

    @QtCore.Slot()
    def updatePosition(self):
        parent = self.parent()
        if parent:
            parent_size = parent.size()
            x = 5
            y = parent_size.height() - self.height() - 5
            self.move(x, y)
            
    
class GL3DViewWidget(opengl.GLViewWidget):    
    sigPrepareForPaint = QtCore.Signal()
    sigViewAngleChanged = QtCore.Signal(object)
    
    def __init__(self,
                 parent : QtCore.QObject | None = None,
                 viewBox: GL3DViewBox | None = None,
                 axisItems: dict[str, GL3DAxisItem] | None = None,
#                 defaultWorldRange:Tuple[Tuple[float, float], ...]|None=None,
                 axis_relative_lengths: Tuple[float, float, float]=(1,1,1),
                 padding=[0, 0, 0],
                 **kwargs
                 ):
        self.vb=None
        self.custom_modifier = QtCore.QKeyCombination(QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier | QtCore.Qt.KeyboardModifier.AltModifier)
        self.key_sec = QtGui.QKeySequence(QtCore.Qt.KeyboardModifier.ControlModifier, QtCore.Qt.KeyboardModifier.ShiftModifier, QtCore.Qt.KeyboardModifier.AltModifier)
        print(parent)
        super().__init__(parent=parent)
        
        defaultWorldRange = ((0.,1.), (0., 1.), (0., 1.))
        
        self.opts.update({"cachedPosition" : None})
        
        self._initView(viewBox, axis_relative_lengths)
        
        self._mouseZoomFactor=0.999
        
        self._zoomingWorld=False
        self._zoomingAxis=False
        self._update_axis_ms=100
        self._update_interval = 50 
        self._last_update_time = 0
        
        self.mouse_info={"mouseAlongAxisPos" : None,
                         "prevmouseAlongAxisPos" : None,
                         "intersectedAxis" : None
                         }
        
        if isinstance(padding, (int, float)):
            padding = [float(padding)]*3
            
        self.axes: Dict[str, List[GL3DAxisItem]] = {i : [] for i in range(3)}
        
        self.opts.update({#"azimuth" : -50,
                          "azimuth" : -45,#45,
                          "distance" : 10,
                          "center" : QtGui.QVector3D(0, 1, 0),
                          #"elevation": -30
                          #"center" : QtGui.QVector3D(0, 0, 0)}
                     #     "elevation" : 0,
                          })
        
        self.opts3D = {"queuedViewUpdate" : False,
                       }
        
        self.setAxisItems(axisItems)

        self.id_map: Dict[uuid4, 'BaseGL3DPlotDataItemMixin'|'CustomPlotDataItem']={}
        self.text_items={ax : {} for ax in "xyz"}
    
        self.autoBtn = GLButtonitem(self, icons.getGraphPixmap('auto'), 15)
        self.autoBtn.setMouseTracking(True)
        self.autoBtn.clicked.connect(self.autoBtnClicked)
      
    def resizeEvent(self, event):
        super().resizeEvent(event)
        if not self.autoBtn is None:
            self.autoBtn.updatePosition()

    @QtCore.Slot()
    def autoBtnClicked(self):
        self.vb.enableAutoRange(enable=True)
        self.vb.sigRangeChangedManually.emit(self.vb.mouseEnabled()[:])

    def _initView(self, viewBox, relative_axis_lengths):
        worldRange = tuple((0, length) for length in relative_axis_lengths)
        if viewBox is None:
            viewBox = GL3DViewBox.GL3DViewBox(worldRange=worldRange)
        self.vb = viewBox
        self.vb.setParentWidget(self)
        super().addItem(self.vb)
        
        self._cameraPosArr = np.array([0., 0., 0.])
        self._centerPosArr = np.array([0., 0., 0.])    

    def initializeGL(self):
        super().initializeGL()
        glEnable(GL_DEPTH_TEST)  

    def create_border(self):
        
        worldRange = self.vb.state["worldRange"]

        corners = np.array([[0, 0, 0], 
                            [1, 0, 0],  
                            [1, 1, 0], 
                            [0, 1, 0]
                            ])
        
        corners = np.array([[worldRange[0][0], worldRange[1][0], worldRange[2][0]],
                           [worldRange[1][1], worldRange[1][0], worldRange[2][0]],
                           [worldRange[1][1], worldRange[1][1], worldRange[2][0]],
                           [worldRange[0][0], worldRange[1][1], worldRange[2][0]]]
                           )

        segments = np.vstack([corners[i] for i in range(4) for _ in (0, 1)])
        
        
      #  segments = [[x, y, z] for z in worldRange[0] for y in worldRange[1] for x in worldRange[2]]
        segments[1::2] = np.roll(corners, -1, axis=0)  
        
        self._borders=[]
        borderline = opengl.GLLinePlotItem(pos=segments,
                                           color=(130,130,130,130),
                                           glOptions="translucent",
                                           mode="lines"
                                           )
        self._borders.append(borderline)
        super().addItem(borderline)
        
    def compute_optimal_clipping(self):
        largest_point_distance = np.sqrt(3)
        near_clip = max(0.0001, self.opts['distance'] - 1.01 * largest_point_distance) 
        far_clip = self.opts['distance'] 
        return near_clip, far_clip

    def prepareForPaint(self):     
        self.sigPrepareForPaint.emit()
        return 
    
    def paintGL(self):
        self.prepareForPaint()
        return super().paintGL()
                    
    def addItem(self, item, ignoreBounds=False):
        if ignoreBounds:
            super().addItem(item)
        else:
            if not self.vb is None:
                self.vb.addItem(item, ignoreBounds=ignoreBounds)
            else:
                super().addItem(item)
            
    def removeItem(self, item):
        if item.parentItem() == self.vb.childGroup:
            self.vb.removeItem(item)
        else:
            super().removeItem(item)

    def setAxisItems(self, axisItems: 'List[GL3DAxisItem] | None' = None) -> None:
        if axisItems is None:
            axisItems = []
        
        for axis in axisItems:
            origin = axis.geometry["worldOrigin"]
            direction = axis.geometry["axis"]
            
            if direction in self.axes:
                for oldAxis in self.axes[direction]:
                    if all(((np.array(oldAxis.geometry["worldOrigin"]) == origin).all(),
                            (oldAxis.geometry["axisFace"] == np.array(axis.geometry["axisFace"])).all()
                            )):
                        oldAxis.unlinkFromView()
                        self.removeItem(oldAxis)
                        self.axes[direction].remove(oldAxis)

            self.axes[direction].append(axis)
            axis.linkToView(self.vb)
            super().addItem(axis)

    def showAxis(self, axis_direction_coords: Tuple[int, Tuple[float, float, float]], show: bool=True):
        s = self.getScale(axis_direction_coords)
        """
        if show:
            s.show()
        else:
            s.hide()
        """

    def getScale(self, axis_direction_coords: Tuple[Tuple[float, float, float], Tuple[int, int]]):
        return self.getAxis(*axis_direction_coords)
    
    def getAxis(self, direction: int, axis_specs: Tuple[Tuple[float, float, float], Tuple[int, int]]=None) -> List[GL3DAxisItem]:
        """
        Return the specified AxisItem.

        Parameters
        ----------
        direction : {0, 1, 2}
            Axis direction of the axis to return.
            
        orthogonalVertices : If origin O of is 0,0,0 of the unit-view, then the x-axis with plane (xy) that goes along x: 0 to 1 with z==0
                                would have the orthogalVertices: {'direction' : (1, 0, 0),
                                                                  'planeDirection' : (0, 1, 0),
                                                                  'perpendicularDirection : (0, 0, 1)
                                                                  }
                                

 
        Returns
        -------
        AxisItem
            The :class:`~GLAxisItem`.

        Raises
        ------
        KeyError
            If the specified axis is not present.
        """
        if direction in self.axes:
            if axis_specs is None:
                return self.axes[direction]
            for ax in self.axes[direction]: 
                if all([p1 == p2 for p1, p2 in zip(ax.geometry["worldOrigin"], axis_specs[0])]) and all([p1 == p2 for p1, p2 in enumerate(axis_specs[1]) if p1 != ax.geometry["axisOrtho"]]):
                    return ax
        raise ValueError(f"Axis with direction {direction} and vertices {axis_specs} is not in the widget")
    
    
    def mousePressEvent(self, ev):
        super().mousePressEvent(ev)
        self.vb.mousePressEvent(ev)

    def mouseReleaseEvent(self, ev):
        self.vb.mouseReleaseEvent(ev)

    def mouseMoveEvent(self, ev):
        if self.vb.userInteracting():
            self.vb.mouseMoveEvent(ev)
        else:
            super().mouseMoveEvent(ev)
        
    @classmethod
    def zoom_2d(cls, point, drag, viewRange, base=1.1):
        axis_mins = np.array([vr[0] for vr in viewRange])
        axis_maxs = np.array([vr[1] for vr in viewRange])

        scales = base ** -drag
        
        new_axis_mins = scales * axis_mins + (1 - scales) * point
        new_axis_maxs = scales * axis_maxs + (1 - scales) * point
        return [[new_min, new_max] for new_min, new_max in zip(new_axis_mins, new_axis_maxs)]


    def cameraPositionNumpy(self):
        p = self.cameraPosition()
        self._cameraPosArr[:] = p.x(), p.y(), p.z()
        return self._cameraPosArr
    
    def centerPositionNumpy(self):
        p = self.opts['center']
        self._centerPosArr[:] = p.x(), p.z(), p.z()
        return self._centerPosArr


    @classmethod
    def map_2D_coords_to_3D(cls, widget: GL3DViewWidget, x: float, y: float):
        widget_width = widget.width()
        widget_height = widget.height()
        device_pixel_ratio = widget.window().screen().devicePixelRatio()

        ndc_x = x / widget_width
        ndc_y = y / widget_height

        viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
        _, _, viewport_width, viewport_height = viewport

        mouse_x_physical = ndc_x * viewport_width
        mouse_y_physical = ndc_y * viewport_height
        mouse_y_physical = viewport_height - mouse_y_physical 
        
        depth = GL.glReadPixels(int(mouse_x_physical), int(mouse_y_physical), 1, 1, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT)[0][0]

        modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
        projection = np.array(widget.projectionMatrix(viewport, (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio* widget_height)).data()).reshape(4, 4)
        
        world_x, world_y, world_z = GLU.gluUnProject(mouse_x_physical, mouse_y_physical, depth, modelview, projection, viewport)

        return world_x, world_y, world_z
    
    def cameraPosition(self, opts=None):
        if opts is None:
            cachedPosition = self.opts["cachedPosition"]
            if cachedPosition is None:
                return super().cameraPosition()
            else:
                return cachedPosition
        else:
            "Getting camera position given opts"
            old_opts = self.opts.copy()
            self.opts = opts
            pos = super().cameraPosition()
            self.opts = old_opts
            return pos

    def cameraTuple(self):
        center = self.opts['center']
        dist = self.opts['distance']
        if self.opts['rotationMethod'] == "quaternion":
            camera = center - self.opts['rotation'].rotatedVector(Vector(0, 0, dist) )
            return camera.x(), camera.y(), camera.z()
        else:
            elev = math.radians(self.opts['elevation'])
            azim = math.radians(self.opts['azimuth'])
            
        return center.x() + dist * math.cos(elev) * math.cos(azim), center.y() + dist * math.cos(elev) * math.sin(azim), center.z() + dist * math.sin(elev)

    def points_intersect_axis(size, start, end):
        
        ndx = start[0] - end[0]
        ndy = start[1] - end[1]
        ndz = start[2] - end[2]

        sxy = ndx * size[1]
        sxz = ndx * size[2]
        syx = ndy * size[0]
        syz = ndy * size[2]
        szx = ndz * size[0]
        szy = ndz * size[1]

        cxy = end[0]*start[1] - end[1]*start[0]
        cxz = end[0]*start[2] - end[2]*start[0]
        cyz = end[1]*start[2] - end[2]*start[1]

        axy = abs(ndx*ndy)
        axz = abs(ndx*ndz)
        ayz = abs(ndy*ndz)
 
        face_num = 0
        axis=None
        face_tau = abs(ndz*axy)

        if start[0] < 0 and 0 < end[0]:
            tau = -start[0] * ayz
            if tau < face_tau and cxy >= 0 and cxz >= 0 and cxy <= -sxy and cxz <= -sxz:
                face_tau = tau
                face_num = 1
                axis=0

        elif end[0] < size[0] and size[0] < start[0]:
            tau = (start[0] - size[0]) * ayz
            if tau < face_tau and cxy <= syx and cxz <= szx and cxy >= syx - sxy and cxz >= szx - sxz:
                face_tau = tau
                face_num = 2
                axis=0

        if start[1] < 0 and end[1] > 0:
            tau = -start[1] * axz
            if tau < face_tau and cxy <= 0 and cyz >= 0 and cxy >= syx and cyz <= -syz:
                face_tau = tau
                face_num = 3
                axis=1

        elif start[1] > size[1] and end[1] < size[1]:
            tau = (start[1] - size[1]) * axz
            if tau < face_tau and cxy >= -sxy and cyz <= szy and cxy <= syx - sxy and cyz >= szy - syz:
                face_tau = tau
                face_num = 4
                axis=1

        if start[2] < 0 and end[2] > 0:
            tau = -start[2] * axy
            if tau < face_tau and cxz <= 0 and cyz <= 0 and cxz >= szx and cyz >= szy:
                face_tau = tau
                face_num = 5
                axis=2

        elif start[2] > size[2] and end[2] < size[2]:
            tau = (start[2] - size[2]) * axy
            if tau < face_tau and cxz >= -sxz and cyz >= -syz and cxz <= szx - sxz and cyz <= szy - syz:
                face_tau = tau
                face_num = 6
                axis=2

        if face_num > 0:
            return axis
        else:
            return None
        
    def points_intersect_axis_points(size, start, end):
        ndx = start[0] - end[0]
        ndy = start[1] - end[1]
        ndz = start[2] - end[2]

        sxy = ndx * size[1]
        sxz = ndx * size[2]
        syx = ndy * size[0]
        syz = ndy * size[2]
        szx = ndz * size[0]
        szy = ndz * size[1]

        cxy = end[0]*start[1] - end[1]*start[0]
        cxz = end[0]*start[2] - end[2]*start[0]
        cyz = end[1]*start[2] - end[2]*start[1]

        axy = abs(ndx*ndy)
        axz = abs(ndx*ndz)
        ayz = abs(ndy*ndz)
        axyz = abs(ndz*axy)

        face_num = 0
        face_tau = abs(ndz*axy)

        if start[0] < 0 and 0 < end[0]:
            tau = -start[0] * ayz
            if tau < face_tau and cxy >= 0 and cxz >= 0 and cxy <= -sxy and cxz <= -sxz:
                face_tau = tau
                face_num = 1

        elif end[0] < size[0] and size[0] < start[0]:
            tau = (start[0] - size[0]) * ayz
            if tau < face_tau and cxy <= syx and cxz <= szx and cxy >= syx - sxy and cxz >= szx - sxz:
                face_tau = tau
                face_num = 2

        if start[1] < 0 and end[1] > 0:
            tau = -start[1] * axz
            if tau < face_tau and cxy <= 0 and cyz >= 0 and cxy >= syx and cyz <= -syz:
                face_tau = tau
                face_num = 3

        elif start[1] > size[1] and end[1] < size[1]:
            tau = (start[1] - size[1]) * axz
            if tau < face_tau and cxy >= -sxy and cyz <= szy and cxy <= syx - sxy and cyz >= szy - syz:
                face_tau = tau
                face_num = 4

        if start[2] < 0 and end[2] > 0:
            tau = -start[2] * axy
            if tau < face_tau and cxz <= 0 and cyz <= 0 and cxz >= szx and cyz >= szy:
                face_tau = tau
                face_num = 5

        elif start[2] > size[2] and end[2] < size[2]:
            tau = (start[2] - size[2]) * axy
            if tau < face_tau and cxz >= -sxz and cyz >= -syz and cxz <= szx - sxz and cyz <= szy - syz:
                face_tau = tau
                face_num = 6

        if face_num > 0:
            tend = face_tau / axyz
            tstart = 1.0 - tend
            return np.round((tstart*start[0]+tend*end[0], tstart*start[1]+tend*end[1], tstart*start[2]+tend*end[2]),6)
        else:
            return None
        
    
        
    def wheelEvent(self, ev):
        if self.vb.userInteracting():
            self.vb.wheelEvent(ev)
        else:
            super().wheelEvent(ev)
            
    def get_unzoom_transform(self):
        """
        Returns a QMatrix4x4 that would visually undo the current zoom level.
        The inverse of the current zoom transformation.
        """
        # Calculate current zoom factor (assuming distance is inversely proportional to zoom)
        # If distance decreases, zoom increases, and vice versa
        base_distance = 4.0  # Reference distance where zoom=1.0
        current_zoom = base_distance / self.opts['distance']
        unzoom_matrix = QtGui.QMatrix4x4()
        unzoom_matrix.scale(current_zoom / 1.0)
        return unzoom_matrix

    def orbit(self, azim, el):
        if azim != 0 and el!= 0:
            self.sigViewAngleChanged.emit(self)
            self.opts["cachedPosition"]=None
        super().orbit(azim, el)

    
    def _outsideBox(self, opts):
        if any([self.vb.state["worldRange"][idx][1] <= pos for idx, pos in enumerate(self.cameraPosition(opts))]):
            return True
        else:
            return False
        
    def _checkCartesianBounds(self, coords):
        return [self.vb.state["worldRange"][i][0] <= coords[i] < self.vb.state["worldRange"][i][1] for i in range(3)]


    def _emit_view_changed(self):
        self._last_update_time = QtCore.QTime.currentTime().msecsSinceStartOfDay()
        self._pending_update = False
        self.viewChangedSignal.emit(self)
        
                
                
    def _delayed_view_update(self):
        if self._pending_update:
            self._emit_view_changed()
        
    def simulateKeys(self, ctrl=False, shift=False):
        """Programmatically set key states"""
        self._ctrl_pressed = ctrl
        self._shift_pressed = shift
        self.update()

    def keyPressEvent(self, ev):
        self.vb.keyPressEvent(ev)
        super().keyPressEvent(ev)

    def keyReleaseEvent(self, ev):
        self.vb.keyPressEvent(ev)
        super().keyReleaseEvent(ev)

    def focusOutEvent(self, event):
        self.vb.focusOutEvent(event)
        super().focusOutEvent(event)  # Call parent method
    

    def queuedViewUpdate(self): return self.opts3D["queuedViewUpdate"]
    
    
    


import numpy as np
import pyqtgraph.opengl as gl
from PySide6 import QtWidgets, QtGui, QtCore
import time
import sys



def get_item_data(item):
    if isinstance(item, gl.GLScatterPlotItem):
        return item.pos
    elif isinstance(item, gl.GLMeshItem):
        return item.opts['meshdata'].vertexes()
    elif isinstance(item, gl.GLLinePlotItem):
        return item.pos
    elif isinstance(item, gl.GLBarGraphItem):
        return item.points()
    return None

def compute_scene_bbox(items):
    min_vals, max_vals = None, None
    
    for item in items:
        data = get_item_data(item)
        if data is None or len(data) == 0:
            continue
            
        if not isinstance(data, np.ndarray):
            data = np.array(data)
            
        if data.ndim == 1:
            data = data[np.newaxis, :]
        if data.shape[1] == 2:
            data = np.hstack([data, np.zeros((len(data), 1))])
            
        item_min = np.nanmin(data, axis=0)
        item_max = np.nanmax(data, axis=0)
        
        if min_vals is None:
            min_vals, max_vals = item_min, item_max
        else:
            min_vals = np.minimum(min_vals, item_min)
            max_vals = np.maximum(max_vals, item_max)
    
    if min_vals is None:
        return np.array([0, 0, 0]), np.array([1, 1, 1])
    
    return min_vals, max_vals

class AxisItem(gl.GLGraphicsItem.GLGraphicsItem):
    def __init__(self, size=None, color=(255, 255, 255, 76.5), antialias=True, glOptions='translucent', parentItem=None):
        super().__init__()

        self.lineplot = None    
        self.tickValues=None
        self.tickSize=1
        
        if size is None:
            size = QtGui.QVector3D(20,20,1)
        self.setSize(size=size)
        self.setSpacing(1, 1, 1)
        self.setTickColor(color)

        self.lineplot = gl.GLLinePlotItem(
            parentItem=self, glOptions=glOptions, mode='lines', antialias=antialias
        )
        self.setParentItem(parentItem)
        self.updateLines()

    def setSize(self, x=None, y=None, z=None, size=None):
        """
        Set the size of the axes (in its local coordinate system; this does not affect the transform)
        Arguments can be x,y,z or size=QVector3D().
        """
        if size is not None:
            x = size.x()
            y = size.y()
            z = size.z()
        self.__size = [x,y,z]
        self.updateLines()
        
    def size(self):
        return self.__size[:]

    def setSpacing(self, x=None, y=None, z=None, spacing=None):
        if spacing is not None:
            x = spacing.x()
            y = spacing.y()
            z = spacing.z()
        self.__spacing = [x,y,z]
        self.updateLines()
        
    def spacing(self):
        return self.__spacing[:]
        
    def setTickColor(self, color):
        self.__tickColor = color
        self.updateLines()

    def tickColor(self):
        return self.__tickColor

    def updateLines(self):
        if self.lineplot is None:
            return

        tickPos = np.array([])
        for i in range(1, 10, 10):
            text_item = gl.GLTextItem(pos = (i, 0 , 0), text=f"{i}", color="white")
            tickPos = np.vstack((tickPos, np.array([i, 0 , 0])))
            tickPos = np.vstack((tickPos, np.array([i, self.tickSize , 0])))
            
        self.lineplot.setData(pos=tickPos, color=self.tickColor())
        self.update()
        

    
class ViewBox(gl.GLGraphicsItem.GLGraphicsItem):
    def __init__(self, *args, **kwargs):
        self.padding = 0.3
        self.min_vals, self.max_vals = None, None
        self.vb_min, self.vb_max = None, None
        self.transform_matrix=None
        super().__init__(*args, **kwargs)
        
    def normalize_scene(self):
        min_vals, max_vals = compute_scene_bbox(self.childItems())
        if self.min_vals is None or any(min_vals < self.vb_min) or any(max_vals > self.vb_max):
            self.min_vals = min_vals
            self.max_vals = max_vals
            
            diff = max_vals - min_vals
            

            self.vb_min = min_vals - diff * self.padding
            self.vb_max = max_vals + diff* self.padding
            
            ranges = self.vb_max - self.vb_min
            ranges[ranges == 0] = 1.0
            
            self.transform_matrix = QtGui.QMatrix4x4()
        
            self.transform_matrix.scale(1/ranges[0], 1/ranges[1], 1/ranges[2])
            self.transform_matrix.translate(-self.vb_min[0], -self.vb_min[1], -self.vb_min[2])
            for child in self.childItems():
                child.setTransform(self.transform_matrix)

        
class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        central = QtWidgets.QWidget()
        self.setCentralWidget(central)
        self.central_layout = QtWidgets.QVBoxLayout()
        central.setLayout(self.central_layout)
        self.gl_view = gl.GLViewWidget()
        self.parent_item = ViewBox()
        self.gl_view.addItem(self.parent_item)
    
    def create_axis(self):
        self.axis_item = AxisItem()
        
        
        
        
    def create_surface(self):
        X = np.arange(-5, 5, 0.5)
        Y = np.arange(-5, 5, 0.5)
        X_mat, Y_mat = np.meshgrid(X, Y)
        
        R_shifted = np.sqrt((X_mat)**2 + (Y_mat)**2)
        Z_mat = 40 + np.sin(R_shifted)                
        X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
        
        pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
        pos_surface = [X, Y, Z_mat]

        
        scatter = gl.GLScatterPlotItem(pos=pos_scatter, color=(1, 0, 0, 1))
        surface = gl.GLSurfacePlotItem(*pos_surface,
                                    glOptions='opaque',
                                    color=(0.5, 0.5, 1, 1))
        scatter.setParentItem(self.parent_item)
        surface.setParentItem(self.parent_item)
        self.scatter=scatter
        self.surface=surface
        self.parent_item.normalize_scene()
        
        self.gl_view.show()
        

        ax = gl.GLAxisItem()
        ax.setSize(1,1,1)
        self.gl_view.addItem(ax)

        
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(500)
        
    def update(self):
        i,j=np.unravel_index(np.argmax(self.surface._z), self.surface._z.shape)
        z_new= self.surface._z.copy()
        z_new[i,j] = z_new[i,j]+0.5
        self.surface.setData(z=z_new)
        
        pos = self.scatter.pos.copy()
        idx = np.argmax(pos[:,0])
        pos[idx,0] = pos[idx,0]+0.1
        self.scatter.setData(pos=pos)
        self.parent_item.normalize_scene()

  #  view.addItem(scatter)
  #  view.addItem(surface)

    
    
def main():
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from .GL3DViewBox import GL3DViewBox
    from enum import Flag 
    from pyvol_terminal.gl_3D_graphing.widgets.GL3DViewWidget import GL3DViewWidget

import weakref
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets, isQObjectAlive
import operator
from functools import reduce
from abc import ABCMeta, abstractmethod
import numpy as np
#from ..import meta
from pyvol_terminal.gl_3D_graphing import meta
from pyqtgraph import Transform3D
from pyqtgraph import functions as fn
import abc
import traceback
import inspect
import warnings
from pyqtgraph.Qt import QT_LIB
from PySide6.QtWidgets import QGraphicsItem
from PySide6 import QtCore
from OpenGL import GL
from functools import wraps

from enum import Enum, EnumType, Flag, IntFlag

from pyqtgraph.Qt import QtCore, QtGui

GLOptions = {
    'opaque': {
        GL.GL_DEPTH_TEST: True,
        GL.GL_BLEND: False,
        GL.GL_CULL_FACE: False,
    },
    'translucent': {
        GL.GL_DEPTH_TEST: True,
        GL.GL_BLEND: True,
        GL.GL_CULL_FACE: False,
        'glBlendFuncSeparate': (GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA,
                                GL.GL_ONE, GL.GL_ONE_MINUS_SRC_ALPHA),
    },
    'additive': {
        GL.GL_DEPTH_TEST: False,
        GL.GL_BLEND: True,
        GL.GL_CULL_FACE: False,
        'glBlendFunc': (GL.GL_SRC_ALPHA, GL.GL_ONE),
    },
}    




qt_enums = [QGraphicsItem.GraphicsItemChange.ItemPositionChange,
            QGraphicsItem.GraphicsItemChange.ItemPositionHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemVisibleChange, 
            QGraphicsItem.GraphicsItemChange.ItemVisibleHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemSceneChange, 
            QGraphicsItem.GraphicsItemChange.ItemTransformChange,
            QGraphicsItem.GraphicsItemChange.ItemTransformHasChanged,
            QGraphicsItem.GraphicsItemChange.ItemSceneHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemParentChange,
            QGraphicsItem.GraphicsItemChange.ItemParentHasChanged,
            QGraphicsItem.GraphicsItemChange.ItemChildAddedChange,
            QGraphicsItem.GraphicsItemChange.ItemChildRemovedChange,
            QGraphicsItem.GraphicsItemChange.ItemFlagsChange,
            QGraphicsItem.GraphicsItemChange.ItemFlagsHaveChanged
            ]
    

class GLGraphicsEnumMetaclass(EnumType):
    def __new__(metacls, clsname, bases, clsdict):
        base_enums = clsdict.get('_base_enums', [])
        custom_members = clsdict.get('_custom_members', [])

        ignore_list = list(clsdict.get('_ignore_', []))
        ignore_list.extend(['_base_enums', '_custom_members'])
        clsdict['_ignore_'] = ignore_list

        values = [member.value for member in base_enums]
        max_value = max(values) if values else 0

        for member in base_enums:
            clsdict[member.name] = member.value

        shifts = [1, 2]
        for idx, (name, _) in enumerate(custom_members):
            shift = shifts[idx] if idx < len(shifts) else (idx + 1)
            value = max_value << shift
            clsdict[name] = value

        return super().__new__(metacls, clsname, bases, clsdict)
    
    def __call__(cls, value):
        return super().__call__(value)


qt_flags = [QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations,
            QGraphicsItem.GraphicsItemFlag.ItemHasNoContents,
            QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges,
            ]


class MetaFlag(type(Flag)):
    def __call__(cls, value=0):
        return super().__call__(value)

class AbstractGLGraphicsItem(QtCore.QObject, metaclass=meta.QABCMeta):
    
    _nextId = 0
    
        
    class GLGraphicsItemChange(Enum, metaclass=GLGraphicsEnumMetaclass):
        _ignore_ = ['_base_enums', '_custom_members'] 
        _base_enums = qt_enums
        _custom_members = [('ItemViewChange', None), ('ItemViewHasChanged', None),
                           ('ItemParentWidgetChange', None), ("ItemParentWidgetHasChanged", None),
          #                 ('ItemDataChange', None), ("ItemParentWidgetHasChanged", None),
                           ]

    class GLGraphicsItemFlag(Flag, metaclass=MetaFlag):
        NoFlag = 0
        ItemIgnoresTransformations = QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations.value
        ItemHasNoContents = QGraphicsItem.GraphicsItemFlag.ItemHasNoContents.value
        ItemSendsGeometryChanges = QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges.value


        def __repr__(self):
            if self.value == 0:
                return f"{self.__class__.__name__}({self.value})"
            return super().__repr__()

    
    wrapped_methods = ["setData", "paint"]
    

    def __init__(self, parentItem: 'AbstractGLGraphicsItem'=None, **kwargs):
        self.__parent: AbstractGLGraphicsItem | None = None
        self.__view = None
        self.__children: list[AbstractGLGraphicsItem] = []#list()
        self.__transform = Transform3D()
        self.__visible = True
        self.__flags = self.GLGraphicsItemFlag()
        self.__initialized = False
        self.__glOpts = {}
        self._connectedView = None
        self._id = AbstractGLGraphicsItem._nextId
        AbstractGLGraphicsItem._nextId += 1
        self.__blockUpdates: bool=False
        self.__cachedView=None
        self.__parentWidget=None

        super().__init__()
       # self.paintClass
        
        

        
        self.blockUpdates(False)
        self.setFlag(self.GLGraphicsItemFlag.ItemSendsGeometryChanges)
        self.setParentItem(parentItem)
        self.setDepthValue(0)
        

                
    def blockUpdates(self, flag):
        self.__blockUpdates=flag

    def setFlag(self, flag, enabled=True):
        if isinstance(flag, Flag):
            _ = self.itemChange(self.GLGraphicsItemChange.ItemFlagsChange, enabled)
            
            if enabled:
                new_flags = self.flags() | flag
            else:
                new_flags = self.flags() & ~flag
            
            if self.flags() != new_flags:
                self.__flags = new_flags
            _ = self.itemChange(self.GLGraphicsItemChange.ItemFlagsHaveChanged, enabled)
        
        
    def flags(self):    
        return self.__flags

    def setParentItem(self, item: 'AbstractGLGraphicsItem'):
        changed_enums=[]
        
        if not self.__parent is None:
            if item == self.__parent:
                warnings.warn(f"{item.__class__.__name__} is already a parentItem to {self.__class__.__name__}, ignoring...", UserWarning)
                return
            else:
                self.itemChange(self.GLGraphicsItemChange.ItemParentChange, item)
                self.itemChange(self.GLGraphicsItemChange.ItemChildRemovedChange, item)
                changed_enums.append((self.GLGraphicsItemChange.ItemParentHasChanged, item))
                self.__parent.__removeChild(self)
            
        if not item is None:    
            if len(changed_enums) == 0:
                self.itemChange(self.GLGraphicsItemChange.ItemParentChange, item)
                changed_enums.append((self.GLGraphicsItemChange.ItemParentHasChanged, item))
            changed_enums.append((self.GLGraphicsItemChange.ItemChildAddedChange, self))
            item.__addChild(self)
            
        if not self.__view is None:
            self.itemChange(self.GLGraphicsItemChange.ItemViewChange, item)
            changed_enums.append((self.GLGraphicsItemChange.ItemViewHasChanged, item))
            self.__view.removeItem(self)
        self.__parent = item
        self.__view = None
        
        for change, value in changed_enums:
            self.itemChange(change, value)
    
    
    def __addChild(self, childItem: 'AbstractGLGraphicsItem'):
        """Do NOT call this method or override it with setParentItem to the childItem. 
        Doing so will cause infinite recursion."""
        self.__children.append(childItem)

        
    def __removeChild(self, childItem: 'AbstractGLGraphicsItem'):
        if childItem in self.__children:
            self.__children.remove(childItem)
    

    def visible(self):
        return self.__visible
        
    def itemChange(self, change, value):
        if change in [self.GLGraphicsItemChange.ItemParentHasChanged, self.GLGraphicsItemChange.ItemSceneHasChanged]:
            if self.__class__.__dict__.get('parentChanged') is not None:
                # user's GraphicsObject subclass has a parentChanged() method
                warnings.warn(
                    "parentChanged() is deprecated and will be removed in the future. "
                    "Use changeParent() instead.",
                    DeprecationWarning, stacklevel=2
                )
                if QT_LIB == 'PySide6' and QtCore.__version_info__ == (6, 2, 2):
                    # workaround PySide6 6.2.2 issue https://bugreports.qt.io/browse/PYSIDE-1730
                    # note that the bug exists also in PySide6 6.2.2.1 / Qt 6.2.2
                    getattr(self.__class__, 'parentChanged')(self)
                else:
                    self.parentChanged()
            else:
                self.changeParent()
        return value

    
    
    def setGLOptions(self, opts):

        if isinstance(opts, str):
            opts = GLOptions[opts]
        self.__glOpts = opts.copy()
        self.update()
        
    def updateGLOptions(self, opts):
        """
        Modify the OpenGL state options to use immediately before drawing this item.
        *opts* must be a dictionary as specified by setGLOptions.
        Values may also be None, in which case the key will be ignored.
        """
        self.__glOpts.update(opts)
        
    
    def parentItem(self):
        """Return a this item's parent in the scenegraph hierarchy."""
        return self.__parent
        
        
    def _setView(self, v):
        self.__view = v
        
    def view(self):
        if self.__parent is None:
            # top level object
            return self.__view
        else:
            # recurse
            return self.__parent.view()
        
    def setDepthValue(self, value):

        self.__depthValue = value
        
    def depthValue(self):
        return self.__depthValue
    
    def __setTransformHelper(self, tr):
        self.itemChange(self.GLGraphicsItemChange.ItemTransformChange, tr)
        self.__transform = tr
        self.itemChange(self.GLGraphicsItemChange.ItemTransformHasChanged, tr)
        self.update()

        
    def setTransform(self, tr):
        if not isinstance(tr, Transform3D):
            tr = Transform3D(tr)
        if tr != self.transform():
            self.__setTransformHelper(tr)
        
    def resetTransform(self):
        if self.transform().isIdentity():
            return 
        self.__setTransformHelper(Transform3D())

        
    def applyTransform(self, tr, local):
        
        if local:
            newTransform = self.transform() * tr
        else:
            newTransform = tr * self.transform()
        if newTransform != self.transform():
            self.__setTransformHelper(newTransform)

            
    def transform(self) -> Transform3D:
        """Return this item's transform object."""
        return self.__transform
        
    def viewTransform(self):
        tr = self.__transform
        p = self
        while True:
            p = p.parentItem()
            if p is None:
                break
            tr = p.transform() * tr
        return Transform3D(tr)
        
    def translate(self, dx, dy, dz, local=False):
        tr = Transform3D()
        tr.translate(dx, dy, dz)
        self.applyTransform(tr, local=local)
        
    def rotate(self, angle, x, y, z, local=False):
        """
        Rotate the object around the axis specified by (x,y,z).
        *angle* is in degrees.
        
        """
        tr = Transform3D()
        tr.rotate(angle, x, y, z)
        self.applyTransform(tr, local=local)
    
    def scale(self, x, y, z, local=True):
        """
        Scale the object by (*dx*, *dy*, *dz*) in its local coordinate system.
        If *local* is False, then scale takes place in the parent's coordinates.
        """
        tr = Transform3D()
        tr.scale(x, y, z)
        self.applyTransform(tr, local=local)
    
    
    def hide(self):
        self.setVisible(False, True)
        
    def show(self):
        self.setVisible(True, True)
    
    
    def initialize(self):
        self.initializeGL()
        self.__initialized = True

    def isInitialized(self):
        return self.__initialized
    
    def initializeGL(self):
        pass
    
    def setupGLState(self):
        for k,v in self.__glOpts.items():
            if v is None:
                continue
            if isinstance(k, str):
                func = getattr(GL, k)
                func(*v)
            else:
                if v is True:
                    GL.glEnable(k)
                else:
                    GL.glDisable(k)
    
      
    #@abstractmethod
    #def _childPaint(self):...

    
    #@abstractmethod
    #def _paintHelper(self):... 
    
    def paint(self):
        if not (AbstractGLGraphicsItem.GLGraphicsItemFlag.ItemHasNoContents & AbstractGLGraphicsItem.flags(self)): 
            self._paintHelper()
            self._childPaint()

    def update(self):
        if self.updatesBlocked():
            return

        v = self.view()
        if v is None:
            return
        v.update()
    
    def updatesBlocked(self):
        return self.__blockUpdates

    def setVisible(self, value, explicitly):
        
        self.__setVisibleHelper(value, True, explicitly)

    def __setVisibleHelper(self, visible, update, explicitly):
        if explicitly:
            self.__explicitlyHidden = visible 
            
        if not self.parentItem() is None and visible and not self.parentItem().visible():
            return 

        if visible == self.visible():
            return 
        newVisible = self.itemChange(self.GLGraphicsItemChange.ItemVisibleChange, visible)
        if self.visible() == newVisible:
            return
        self.__visible = newVisible
        
        for child in self.childItems():
            if not visible or not child.__explicitlyHidden:
                child.__setVisibleHelper(visible, False, False)
        
        self.update()
        self.itemChange(self.GLGraphicsItemChange.ItemVisibleHasChanged, self.visible())
    
    def childItems(self) -> List['AbstractGLGraphicsItem']:
        return list(self.__children)
    
   
    def mapToParent(self, point):
        tr = self.transform()
        if tr is None:
            return point
        return tr.map(point)
        
    def mapFromParent(self, point):
        tr = self.transform()
        if tr is None:
            return point
        return tr.inverted()[0].map(point)
        
    def mapToView(self, point):
        tr = self.viewTransform()
        if tr is None:
            return point
        return tr.map(point)
        
    def mapFromView(self, point):
        tr = self.viewTransform()
        if tr is None:
            return point
        return tr.inverted()[0].map(point)

    def modelViewMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentModelView()

    def projectionMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentProjection()

    def mvpMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentProjection() * view.currentModelView()


    def _updateView(self):
        ## called to see whether this item has a new view to connect to
        ## NOTE: This is called from GraphicsObject.itemChange or GraphicsWidget.itemChange.

        if not hasattr(self, '_connectedView'):
            # Happens when Python is shutting down.
            return
        self.forgetViewBox()
        
        ## check for this item's current viewbox
        view = self.getViewBox()
        
        oldView = None
        if not self._connectedView is None:
            oldView = self._connectedView()
        
        if view is oldView:
            # "  already have view", view
            return

        ## disconnect from previous view
        if not oldView is None:
            for signal, slot in [('sigRangeChanged', self.viewRangeChanged),
                                 ('sigTransformChanged', self.viewTransformChanged)]:
                try:
                    getattr(oldView, signal).disconnect(slot)
                except (TypeError, AttributeError, RuntimeError):
                    # TypeError and RuntimeError are from pyqt and pyside, respectively
                    pass
            
            self._connectedView = None
            
        if not view is None:
            view.sigRangeChanged.connect(self.viewRangeChanged)
            view.sigTransformChanged.connect(self.viewTransformChanged)
            self._connectedView = weakref.ref(view)
            self.viewRangeChanged()
            self.viewTransformChanged()
        self._replaceView(oldView)        
        self.viewChanged(view, oldView)        

    def setParentWidget(self, parent=None):
        
        if self.parentWidget() != parent:
            self.itemChange(self.GLGraphicsItemChange.ItemParentWidgetChange, parent) 
            self.__parentWidget = parent
            self.itemChange(self.GLGraphicsItemChange.ItemParentWidgetHasChanged, parent) 
    
    def parentWidget(self) -> GL3DViewWidget:
        return self.__parentWidget

    def viewChanged(self, view, oldView):
        """Called when this item's view has changed
        (ie, the item has been added to or removed from a ViewBox)"""
        
    def viewRect(self):
        """Return the visible bounds of this item's ViewBox or GraphicsWidget,
        in the local coordinate system of the item."""
        if self.__cachedView is not None:
            return self.__cachedView

        # Note that in cases of early returns here, the view cache stays empty (None).
        view = self.getViewBox()
        if view is None:
            return None
        bounds = self.mapRectFromView(view.viewRect())
        if bounds is None:
            return None

        bounds = bounds.normalized()
        
        self.__cachedView = bounds
        
        ## nah.
        #for p in self.getBoundingParents():
            #bounds &= self.mapRectFromScene(p.sceneBoundingRect())

        return bounds
        
    def _replaceView(self, oldView, item=None):
        if item is None:
            item = self
        for child in item.childItems():
            if isinstance(child, 'AbstractGLGraphicsItem'):
                if child.getViewBox() is oldView:
                    child._updateView()
                        #self._replaceView(oldView, child)
            else:
                self._replaceView(oldView, child)
        

    def parentChanged(self):
        self.changeParent(self)
        
    def changeParent(self):
        self._updateView()

    
    @QtCore.Slot()
    def viewTransformChanged(self):
        self.__cachedView = None
        
    
        
    def informViewBoundsChanged(self):
        view = self.getViewBox()
        if view is not None and hasattr(view, 'implements') and view.implements('ViewBox'):
            view.itemBoundsChanged(self)
    

    def allChildItems(self, root=None):
        """Return list of the entire item tree descending from this item."""
        if root is None:
            root = self
        tree = []
        for ch in root.childItems():
            tree.append(ch)
            tree.extend(self.allChildItems(ch))
        return tree
    
    @QtCore.Slot(object, object)
    @QtCore.Slot(object, object, object)
    def viewRangeChanged(self, vb=None, ranges=None, changed=None):...


    def mapRectFromView(self, obj):
        vt = self.viewTransform()
        if vt is None:
            return None
        vt = fn.invertQTransform(vt)
        return vt.mapRect(obj)

    def getViewBox2(self):
        

        if self._viewBox is None:
            while True:
                parent = self.parentItem()
                if isinstance(parent, GL3DViewBox):
                    self._viewBox = parent
                    break
            return self._viewBox if self._viewBox else None

    def getViewBox(self) -> GL3DViewBox:
        if self._viewBox is None:
            p = self
            while True:
                try:
                    p = p.parentItem()
                except RuntimeError:  ## sometimes happens as items are being removed from a scene and collected.
                    return None
                if p is None:
                    return None
                if hasattr(type(p), 'implements'):
                    if p.implements('ViewBox'):
                        self._viewBox = weakref.ref(p)
                        break
        
        return self._viewBox() if self._viewBox else None
    
    
    def forgetViewBox(self):
        self._viewBox = None



class _Temp(AbstractGLGraphicsItem):
    pass

class ABCGraphicsItemMeta(meta.QABCMeta):
    def __call__(cls, *args, **kwargs):
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        self = cls.__new__(cls, *args, **kwargs)
        self._original_class = cls
        self.__class__ = _Temp
        parentItem = kwargs.pop("parentItem", None)
        AbstractGLGraphicsItem.__init__(self, parentItem=parentItem)
        
        self.__class__ = self._original_class
        del self._original_class
        
        if hasattr(self, '__init__') and self.__init__ is not None:
            self.__init__(*args, **kwargs)
            
        return self





def get_settings_stylesheets(style_sheet):
    match style_sheet:
        case "QPushButton":
            style_sheet_str = """
                            QPushButton {background-color: #464646;
                                        color: white;
                                        border: 1px solid black;
                                        padding: 5px;
                                        }
                            QPushButton:checked {
                                                background-color: #9e9e9f;
                                                color: black;
                                                }
                                                
                            QPushButton::hover {
                                                color: black;
                                                background-color: #9e9e9f;
                                                }
                            """
                            
                            
        case "QComboBox":
            style_sheet_str = """
                        QComboBox {
                            background-color: #464646;
                            color: white;
                        }
                        QComboBox::drop-down {
                            border-color: #464646;
                            background-color: #464646;
                        }
                        QComboBox::item {
                            background-color: #464646;
                            color: white; 
                        }
                        QComboBox::item:selected {
                            background-color:  #9e9e9f;
                            color: black; 
                        }
                        QComboBox QAbstractItemView {
                            background-color: #464646;
                            color: white; 
                        }
                        """
    
        case "QMenu":
            style_sheet_str = """
                            QMenu::item {
                                        background-color: #464646;
                                        color: white
                                        }
                            QMenu::item:selected {
                                                background-color: #9e9e9f; 
                                                color: black;
                                                }
                            
                            QMenu::hover {color: black;
                                        background-color: #9e9e9f;
                                        }

                            """

        case "QToolButton":
            style_sheet_str = """
                            QToolButton {
                                        color: white;
                                        background-color: #464646;
                                        }
                                        
                            QToolButton::selected {
                                                    color: black;
                                                    background-color: #9e9e9f;
                                                    }

                            QToolButton::hover {
                                                color: black;
                                                background-color: #9e9e9f;
                                                }

                            """

    return style_sheet_str


from PySide6 import QtCore, QtWidgets
import sys

class InnerWorker(QtCore.QObject):
    aslot = QtCore.Signal()
    def __init__(self, name):
        super().__init__()
        self.name = name
        self._thread = QtCore.QThread()  # Each worker has its own thread
        self.aslot.connect(triggered)
    
    def run(self):
        print(f"[{self.name}] running in", QtCore.QThread.currentThread())
        self.aslot.emit()
        
        
def triggered():
    print(f"triggered: {QtCore.QThread.currentThread()}")
    import sys
    sys.exit()

class Manager(QtCore.QObject):
    def __init__(self):
        super().__init__()
        self.worker1 = InnerWorker("Worker 1")
        self.worker2 = InnerWorker("Worker 2")

    def start_inner_threads(self):
        self.worker1.moveToThread(self.worker1._thread)
        self.worker1._thread.started.connect(self.worker1.run)
        self.worker1._thread.start()

        # Start Worker 2 in its own thread  
        self.worker2.moveToThread(self.worker2._thread)
        self.worker2._thread.started.connect(self.worker2.run)
        self.worker2._thread.start()

        print("[Manager] started workers from", QtCore.QThread.currentThread())

class Bootstrap(QtCore.QObject):
    def __init__(self):
        super().__init__()
        self._thread = QtCore.QThread()
        self.manager = Manager()
        self.manager.moveToThread(self._thread)
        self._thread.started.connect(self.manager.start_inner_threads)
        self._thread.start()

app = QtWidgets.QApplication(sys.argv)
b = Bootstrap()
sys.exit(app.exec())


from PySide6.QtWidgets import QApplication, QGraphicsView, QGraphicsScene, QGraphicsRectItem
from PySide6.QtCore import QRectF
import sys
from PySide6.QtOpenGLWidgets import QOpenGLWidget
from pyqtgraph import opengl

app = QApplication([])

view = QGraphicsView()
gl_view = opengl.GLViewWidget()
view.setViewport(gl_view)  # 🔧 Use OpenGL to render the QGraphicsView

scene = QGraphicsScene()
view.setScene(scene)

rect_item = QGraphicsRectItem(QRectF(0, 0, 100, 100))
scene.addItem(rect_item)

view.show()
sys.exit(app.exec())


from PySide6 import QtCore, QtGui



def sheet(opt):
    match opt:
        case "QComboBox":
            style_sheet_str = """
                            QComboBox {
                                background-color: #fb8b1e;
                                color: black;
                                border: None;
                                text-align: center;
                            }
                            QComboBox:focus {
                                background-color: #fb8b1e;
                                border: 2px solid black;
                                text-align: center;
                            }
                            QComboBox::drop-down {
                                border-color: black;
                                background-color: #fb8b1e;
                                text-align: center;
                            }
                            QComboBox::item {
                                background-color: #fb8b1e;
                                color: black;
                                text-align: center;
                            }
                            QComboBox::item:selected {
                                background-color: #d97d1a; 
                                color: black;
                                text-align: center;
                            }
                            QComboBox QAbstractItemView {
                                background-color: #fb8b1e;
                                color: black;
                                border: None;
                                text-align: center;
                            }
                        """
        case "QTableWidget":
            style_sheet_str = """
                QTableView {
                    border-style: none;
                    background-color: black;
                    padding: 0px;
                    margin: 0px;
                    color: white;
                    border-left: 2px solid #414141;
                    border-right: 2px solid #414141;
                    selection-background-color: #9e9e9f;
                }
                QTableWidget::item {
                    padding: 0px;
                    margin: 0px;
                    border: none; 
                }
                QTableWidget::item:selected {
                    background-color: #414141;
                    padding: 0px;
                    margin: 0px;
                    color: white;
                }
            """                   

                        
                        
                        
    return style_sheet_str


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any, Set
if TYPE_CHECKING:
    from pyvol_terminal.utils import AbstractValidPriceChecker
    from engines.yield_engines.implied_engines import AbstractYieldEngine 


import numpy as np
import math
from abc import ABC, abstractmethod
from scipy.stats import norm
from dataclasses import dataclass

import numpy as np
import math
from abc import ABC, abstractmethod
from scipy.stats import norm
from typing import Set, Dict


class ABCEuroEngine(ABC):
    
    _GREEK_REQUIREMENTS = {"delta" : {"d1",
                                     "N_flag_d1"
                                     },
                           "gamma" : {"d1",
                                     "phi_d1"
                                     },
                           "vega" : {"d1",
                                    "phi_d1"
                                    },
                           "theta": {"d1",
                                     "d2",
                                     "phi_d1",
                                     "N_flag_d1",
                                     "N_flag_d2"
                                     },
                           "rho" : {"d2",
                                   "N_flag_d2"
                                   }, 
                           "price": {"d1",
                                     "d2",
                                     "N_flag_d1",
                                     "N_flag_d2"
                                     }
                           }


    
    
    
    
    def __init__(self,
                 strike: float,
                 expiry: float,
                 flag_str: str,
                 flag_int: int=None,
                 interest_rate_engine: AbstractYieldEngine=None,
                 dividend_rate_engine: AbstractYieldEngine=None
                 ):
        self.strike=strike
        self.expiry=expiry
        self.flag=flag_str
        self.flag_str=flag_str
        self.interest_rate_engine=interest_rate_engine
        self.dividend_rate_engine=dividend_rate_engine
        self.IVOL_engine=None
        self.greek_engine_all=None
        self.greeks_engines={}
        self._cachedParams={"t" : np.nan,
                            "px_option" : np.nan,
                            "px_underlying" : np.nan,
                            "r" : np.nan,
                            "q" : np.nan,
                            }
        
        if self.dividend_rate_engine is None:
            self._dividend_off=True
        else:
            self._dividend_off=False
        
        if not flag_int is None:
            self.flag_int = flag_int
        else:
            self.flag_int = 1 if flag_str.lower() == "c" else -1
    
    
    def calculate_IVOL(self, timestamp, px_option, px_underlying): 
        self.internal_calculate_IVOL(timestamp, px_option, px_underlying)
    
    @abstractmethod
    def internal_calculate_IVOL(self, timestamp, price_opt, px_underlying): ...
    
    @abstractmethod
    def calculate_all_greeks(self, timestamp, IVOL, px_underlying): ...

    @abstractmethod
    def calculate_pair_metrics(self, *args, **kwargs): ...

    @abstractmethod
    def set_params(self, *args, **kwargs): ...
    
    def calculate(self,
                  greeks: Set[str],
                  **kwargs,
                  ) -> Dict[str, float]:
        if len(kwargs) >= 0:
            self.set_params(**kwargs)
        cache = self._calculate_intermediates(greeks)
        return self._compute_greeks(greeks, cache)

    @abstractmethod
    def _calculate_intermediates(self, greeks: Set[str]) -> Dict[str, float]: ...

    @abstractmethod
    def _compute_greeks(self, greeks: Set[str], cache: Dict[str, float]) -> Dict[str, float]: ...
    



    



from PySide6.QtWidgets import (
    QApplication, QWidget, QVBoxLayout, QHBoxLayout, 
    QPushButton, QComboBox
)
from PySide6.QtCore import Slot, Qt
from PySide6 import QtWidgets

class Terminal(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        
        main_layout = QVBoxLayout()
        self.setLayout(main_layout)
        
        # Create button layout with 4 buttons
        button_layout = QHBoxLayout()
        self.buttons = []
        for i in range(1, 5):
            button = QPushButton(f"Button {i}")
            button.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
            button.clicked.connect(self.create_button_handler(i))
            self.buttons.append(button)
            button_layout.addWidget(button)
        
        
        self.combo_box = QComboBox()
        self.combo_box.addItems(["No Instrument Selected", "Option 2", "Option 3", "Option 4"])
        self.combo_box.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        self.combo_box.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.combo_box.currentIndexChanged.connect(self.update_button_state)
        
        main_layout.addWidget(self.combo_box)
        main_layout.addLayout(button_layout)
        
        
        self.update_button_state(0)
        self.adjustSize()              # Resize widget to fit contents
        self.setFixedSize(self.size())  # Lock size
            
    def update_button_state(self, index):
        """Enable/disable buttons based on combo box selection"""
        # Disable buttons for first option, enable for others
        state = index != 0
        for button in self.buttons:
            button.setEnabled(state)
    
    def create_button_handler(self, button_id):
        """Create a slot handler for button clicks"""
        @Slot()
        def handler():
            # Get current combo box text
            selected_option = self.combo_box.currentText()
            print(f"Button {button_id} clicked with option: {selected_option}")
            # Add your custom button handling logic here
        return handler

if __name__ == "__main__":
    app = QApplication([])
    
    widget = Terminal()
    widget.setWindowTitle("Combo Button Widget")
    widget.resize(400, 200)
    widget.show()
    
    app.exec()


from __future__ import annotations 
from typing import TYPE_CHECKING

from pyvol_terminal.gl_3D_graphing.widgets import GL3DViewWidget
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GL3DViewBox

from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore
import numpy as np
from OpenGL import GL
from pyqtgraph.Qt import QT_LIB
import importlib
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DAxisItem
from pyvol_terminal.gl_3D_graphing.graphics_items import GL3DViewBox, AbstractGLPlotItem, AbstractGLGraphicsItem
from pyqtgraph import opengl
import sys
QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DGraphicsItems import GL3DSurfacePlotItem, GL3DLinePlotDataItem
from matplotlib import colormaps


def scale_value(x, old_min, old_max, new_min, new_max):
    return new_min + (x - old_min) * (new_max - new_min) / (old_max - old_min)

def normalize(x, p1, p2):
    x_min = np.min(x)
    x_max = np.max(x)
    return (x - x_min) / (x_max - x_min) * (p2 - p1) + p1

class CustomMixedGLSurfacePlotItem(GL3DSurfacePlotItem):
    def __init__(self, colormap, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.colormap = colormap
        
    def _setDataChild2(self, x=None, y=None, z=None, colors=None):        
        if not x is None \
            and not y is None \
                and not z is None \
                    and not colors is None:
            if not self.view() is None:
                vb = self.view().vb
                wz_min, wz_max = vb.opts["worldRange"][2]
                z_norm = scale_value(z, z.min(), z.max(), wz_min, wz_max)
                
                colors = self.colormap(z_norm.flatten()).reshape(*z.shape, 4) 
        return super()._setDataChild(x=x, y=y, z=z, colors=colors)
    
    def _setDataChild(self, x=None, y=None, z=None, colors=None):
        if x is not None and y is not None and z is not None:
            if self.view() is not None:
                vb = self.view().vb
                wz_min, wz_max = vb.state["worldRange"][2]
                # Normalize z-values to [0,1] based on worldRange
                if wz_max != wz_min:
                    z_normalized =  (wz_max - wz_min) / (z - wz_min)
                else:
                    z_normalized = np.zeros_like(z)
                z_normalized = normalize(z, wz_min, wz_max)
                colors = self.colormap(z_normalized.flatten()).reshape(*z.shape, 4)
        return super()._setDataChild(x=x, y=y, z=z, colors=colors)
    
def create_pos(nticks) :
    X = np.linspace(-5, 10, nticks)
    Y = np.linspace(-3, 10, nticks)
    X_mat, Y_mat = np.meshgrid(X, Y, indexing="xy")
    R = np.sqrt(X_mat**2 + Y_mat**2)

    m1 = 20
    
    Z_mat = m1 * np.sin(R) 
    return X_mat, Y_mat, Z_mat, m1

from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DPlotDataItemMixin import BaseGL3DPlotDataItemMixin
from abc import abstractmethod

class ABCPyVolPlotItemMixin(BaseGL3DPlotDataItemMixin):
    
    @abstractmethod
    def getValues(self):...
    
    @abstractmethod
    def id(self):...
    

class PyVolLineItem(ABCPyVolPlotItemMixin, GL3DLinePlotDataItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.item_type = None
        self.dataset = None

    def id(self):
        return self._internal_id
    
    def getValues(self):
        return self.pos[:,0], self.pos[:,1], self.pos[:,2]

def create_objects(X_mat, Y_mat, Z_mat, m1) -> AbstractGLPlotItem:

    
    X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
    
    pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
    """
    scatter = GLScatterPlotItem.GLScatterPlotItem(pos=pos_scatter,
                                                  glOptions='opaque',
                                                  color=(1, 0, 1, 1)
                                                  )
    """
    scatter=None
    pos_surface = X_mat[0], Y_mat[:,1], Z_mat.T
    pos = np.linspace(0, 1, 500)   
    #pre_cmap = pg.colormap.get("inferno")
    

    colormap = colormaps["viridis"]
    
    z_norm = (Z_mat - Z_mat.min()) / (Z_mat.max() - Z_mat.min())
    colors = colormap(z_norm.flatten())
    colors = colors.reshape(*Z_mat.T.shape, 4)


    wz_min, wz_max = 0, 1
    z_norm = scale_value(Z_mat, Z_mat.min(), Z_mat.max(), wz_min, wz_max)
    
    #colors = None  # Let the item handle color mapping dynamically
    
    
    surface = GL3DSurfacePlotItem(
                                           x=pos_surface[0],
                                            y=pos_surface[1],
                                            z=pos_surface[2],
                                            glOptions='opaque',
                                            shader="shaded",
                                            colors=colors
                                            )
    
    surface.setGLOptions("opaque")
    
    line = PyVolLineItem(pos = [[0, 0, 0], [0, 0, 1]])
    line.setGLOptions("opaque")

    return scatter, surface




class DotItem(AbstractGLGraphicsItem.AbstractGLGraphicsItem):
    def __init__(self, *coordinate):
        super().__init__()
        self.coordinate = list(coordinate)
        self.setGLOptions('opaque')
        self.setFlag(self.GLGraphicsItemFlag.ItemIgnoresTransformations)
        
    def initializeGL(self):
        GL.glEnable(GL.GL_POINT_SMOOTH)
        GL.glPointSize(10)
        
    def updateCoordinate(self, coord):
        self.coordinate = coord
        
    def paint(self):
        if not self.view():
            return
            
        self.setupGLState()
        
        # Set up orthographic projection matching view dimensions
        width = self.view().width()
        height = self.view().height()
        
        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()
        GL.glOrtho(0, width, height, 0, -1, 1)
        
        GL.glMatrixMode(GL.GL_MODELVIEW)
        GL.glLoadIdentity()
        
        GL.glColor3f(1.0, 1.0, 0.0)  # White color
        
        GL.glBegin(GL.GL_POINTS)
        GL.glVertex2f(*self.coordinate)  # Use pixel coordinates directly
        GL.glEnd()

class CustomScatter(opengl.GLScatterPlotItem):
    def __init__(self, *args, **kwargs):
        self.coordinate = (300, 300)
        super().__init__(*args, **kwargs)      
        
    

    def paint(self):
        GL.glColor3f(1.0, 1.0, 1.0)
        
        GL.glBegin(GL.GL_POINTS)
        GL.glVertex2f(*(self.coordinate))  
        GL.glEnd()

"""
        ax_orth = worldRange[1][0], worldRange[2][0]
        ax=0
        ax_perp = 1
        axisWidth = abs(worldRange[ax_perp][0] - worldRange[ax_perp][1])
        tickOrientation = -1

        axisItem1 = GLAxisItem(ax,
                               worldRange[ax],
                               ax_orth,   
                               axisWidth, 
                               tickOrientation,
                               ax_perp,
                               text="x-axis"
                               )
        


"""

class Custom(QtWidgets.QGraphicsItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)      

class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, app=None, **kwargs):
        super().__init__(*args, **kwargs)   
        
        worldRange=[[0., 1.5],
                    [-0.7, 1.],
                    [0.,1.]
                    ]
        
        worldRange=[[0., 1.],
                    [0., 1.],
                    [0., 1.]
                    ]

        
        a1 = [[worldRange, 0, (0, 0, 0), 1],
              {"text" : "x-axis",
               "showValues" : True,
               "valuesColor" : "white",
               "labelColor" : "white",
               "faceTickOffset" : -0.1,
               "syncToView" : True,
               "showGrid" : True,
               "showTickMargin" : True,
            #   "showFaceBorder" : False,
               
               }
              ]
        a2 = [[worldRange, 1, (0, 0, 0), 0],
              {"text" : "y-axis",
               "showValues" : True,
               "valuesColor" : "yellow",
               "labelColor" : "yellow",
               "faceTickOffset" : -0.1,
               "syncToView" : True,
               "showGrid" : True,
          #     "showFaceBorder" : False,
               }
              ]
        
        a3 = [[worldRange, 2, (1, 1, 0), 0],
              {"text" : "z-axis",
               "showValues" : True,
               "showTicks" : True,
               "valuesColor" : "cyan",
               "labelColor" : "cyan",
               "faceTickOffset" : -0.1,
               "graphFaceOffset" : -0.3,
               "syncToView" : True,
               "showGrid" : True,
             #  "showFaceBorder" : False,
               }
              ] 
        a33 = [[worldRange, 2, (1, 0, 0), 0],
              {"text" : "z-axis",
               "showValues" : True,
               "showTicks" : True,
               "valuesColor" : "cyan",
               "labelColor" : "cyan",
               "faceTickOffset" : -0.1,
               "graphFaceOffset" : -0.3,
               "syncToView" : True,
               "showGrid" : True,
             #  "showFaceBorder" : False,
               }
              ] 

        a4 = [[worldRange, 2, (0, 0, 0), 1],
              {"text" : "z-axis",
               "showValues" : True,
               "valuesColor" : "cyan",
               "labelColor" : "cyan",
               "faceTickOffset" : -0.1,
               "graphFaceOffset" : -0.3,
               "syncToView" : True,
            #   "showGrid" : False,
           #    "showFaceBorder" : False,

               }
              ]
        a44 = [[worldRange, 2, (0, 1, 0), 1],
              {"text" : "z-axis",
               "showValues" : True,
               "valuesColor" : "cyan",
               "labelColor" : "cyan",
               "faceTickOffset" : -0.1,
               "graphFaceOffset" : -0.3,
               "syncToView" : True,
            #   "showGrid" : False,
           #    "showFaceBorder" : False,

               }
              ]

        a13 = [[worldRange, 0, (0, 1, 0), 2],
              {"text" : "x-axis",
               "showValues" : False,
               "showLabel" : False,
               "valuesColor" : "cyan",
               "labelColor" : "cyan",
               "faceTickOffset" : -0.1,
               "graphFaceOffset" : -0.3,
               "syncToView" : True,
               "showTickMargin" : True,
               }
              ] 
        
        a23 = [[worldRange, 1, (0, 0, 0), 2],
              {"text" : "x-axis",
               "showValues" : False,
               "showLabel" : False,
               "valuesColor" : "cyan",
               "labelColor" : "cyan",
               #"faceTickOffset" : -0.1,
               "graphFaceOffset" : -0.3,
               "syncToView" : True
               }
              ]

        items = [a1, a4]
        items = [a1, a2, a3]

      #  items = [a1, a13]
      #  a1[1]["showGrid"] = False
      #  a1[1]["showFaceBorder"] = False
      #  a2[1]["showFaceBorder"] = False
        
    #    items = [a1, a2]
        
        axis_items = [GL3DAxisItem.fromViewBoxRange(*item[0], **item[1]) for item in items]

     #   axis_items[0].linkOptimizeAxes(axis_items[1])
        
        worldRange=[[-1., 1.],
                    [-1., 1.],[-1., 1.]
                    ]

        worldRange=[[0., 1.],
                    [0., 1.],
                    [0., 1.]
                    ]

        self.viewBox = GL3DViewBox.GL3DViewBox(worldRange=worldRange)
        
        self.gl_view = GL3DViewWidget.GL3DViewWidget(axisItems=axis_items,
                                                 viewBox=self.viewBox
                                                 )
        
                
        self.phase=0
        self.step=0

        self.gl_view.setCameraPosition(distance=5)
        self.setCentralWidget(self.gl_view)
    #    self.setGeometry(0, 0, 800, 600)
        self.update()
        
        self.nticks=50
        
        self.X_mat, self.Y_mat, self.Z_mat, self.m1 = create_pos(self.nticks)
        self.scatter, self.surface = create_objects(self.X_mat, self.Y_mat, self.Z_mat, self.m1)    
        
        self.scatter=None
       # self.gl_view.addItem(self.scatter)
        self.gl_view.addItem(self.surface)
        self.c=0
        
        self.setGeometry(0, 0, 800, 600)
        self.show()
    
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_plots)
        
        for i in range(21):
            if i % 2 == 0:
                txt = opengl.GLTextItem(pos = (i/ 10, i/10, 0), text=str(i))
                #self.gl_view.addItem(txt, ignoreBounds=True)
        
        
        
     #   self.timer.setSingleShot(True)
        


        
    def update_plots(self):
        self.c+=1
        self.phase += 0.4

        growth = 1 + 0.5 * np.sin(self.phase * 0.3)


        step = 0.5
        range_base = 10 * growth
        xmin, xmax = -range_base, range_base
        ymin, ymax = -range_base, range_base

        X = np.linspace(xmin+step, xmax + step, self.nticks)
        Y = np.linspace(ymin+step, ymax + step, self.nticks)
        self.X_mat, self.Y_mat = np.meshgrid(X, Y)
        

        R = np.sqrt(self.X_mat**2 + self.Y_mat**2)

        amplitude = 10 + 5 * np.sin(self.phase * 0.2)
        self.Z_mat = amplitude * np.sin(R - self.phase) * np.sin(R * 0.5 - self.phase)

        X_vec = self.X_mat.flatten()
        Y_vec = self.Y_mat.flatten()
        Z_vec = self.Z_mat.flatten()
        
        pos_scatter = np.column_stack((X_vec, Y_vec, Z_vec))

        
        x_vals = self.X_mat[0]
        y_vals = self.Y_mat[:, 0]
        z_vals = self.Z_mat
        
        x_vals *= 20
        
        y_vals *= 6
        
        self.surface.setData(x=x_vals,
                             y=y_vals , 
                             z=z_vals,
                             )
        
        print(self.gl_view.vb.viewRange())
        
        

def main():
    
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    win.timer.start(50)
    sys.exit(app.exec())

if __name__ == "__main__":
    main()
    
    """
    # Profile the entire application run
    profiler = cProfile.Profile()
    profiler.enable()
    
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    win.timer.start(750)
    app.exec()
    
    profiler.disable()
    
    # Save results to file
    with open('profile_results.txt', 'w') as f:
        ps = pstats.Stats(profiler, stream=f)
        ps.sort_stats('tottime')
        ps.print_stats()
    """


import numpy as np
import time


def convert_unix_maturity_to_years(unix_maturity):
    return (unix_maturity - time.time()) / 3600 / 24 / 365

def convert_to_yte(timestamp, unix_maturity):
    return (unix_maturity - timestamp) / 3600 / 24 / 365

def null_metric(raw_object, x, y, z):
    return x, y, z, [True]*z.size, np.arange(z.size)

def _base_money_sorter(raw_object, x, y, z):
    mask_removal = raw_object.OTM
    x = x[mask_removal]
    y = y[mask_removal]
    z = z[mask_removal]
    
    mask_rearrange = np.lexsort((x, y))
    
    x = x[mask_rearrange]
    y = y[mask_rearrange]
    z = z[mask_rearrange]
    return x, y, z, mask_removal, mask_rearrange

def TVAR_function(raw_object, x, y, z):
    return x, y, z*convert_unix_maturity_to_years(y), [True]*z.size, np.arange(z.size)

def IVOL_perc_function(raw_object, x, y, z):
    return x, y, 100*z, [True]*z.size, np.arange(z.size)

def moneyness_mask_sorter(raw_object, x, y, z):
    x=raw_object.moneyness
    return _base_money_sorter(raw_object, x, y, z)

def log_moneyness_mask_sorter(raw_object, x, y, z):
    x=raw_object.log_moneyness
    return _base_money_sorter(raw_object, x, y, z)

def standardised_moneyness_sorter(raw_object, x, y, z):
    x=raw_object.standardised_moneyness
    return _base_money_sorter(raw_object, x, y, z)

def years_metric_func(raw_object, x, y, z):
    return x, convert_unix_maturity_to_years(y), z, [True]*z.size, np.arange(z.size)

def moneyness_spot(raw_object, x, y, z):
    return raw_object.moneyness, y, z, [True]*z.size, np.arange(z.size)

def log_moneyness_spot(raw_object, x, y, z):
    return raw_object.log_moneyness, y, z, [True]*z.size, np.arange(z.size)

def delta_metric_mask_sorter(raw_object, x, y, z):
    mask_removal = raw_object.OTM & (raw_object.delta_mag < 0.5)
    x = raw_object.delta
    x_masked = x[mask_removal] 
    y_masked = y[mask_removal]  
    z_masked = z[mask_removal]
    
    put_indices = np.where(~raw_object.call_flag[mask_removal])[0]
    call_indices = np.where(raw_object.call_flag[mask_removal])[0]

    if put_indices.size > 0:
        sorted_put_indices = put_indices[np.lexsort((y_masked[put_indices], -x_masked[put_indices]))]
    else:
        sorted_put_indices = np.array([], dtype=int)
    if call_indices.size > 0:
        sorted_call_indices = call_indices[np.lexsort((y_masked[call_indices], -x_masked[call_indices]))]
    else:
        sorted_call_indices = np.array([], dtype=int)

    mask_rearrange = np.concatenate([sorted_put_indices, sorted_call_indices])
    
    x_sorted = x_masked[mask_rearrange]
    y_sorted = y_masked[mask_rearrange]
    z_sorted = z_masked[mask_rearrange]        
    mask = x_sorted > 0
    x_sorted[~mask] = -1 * x_sorted[~mask]  
    x_sorted[mask] = 1 - x_sorted[mask]  
    return x_sorted, y_sorted, z_sorted, mask_removal, mask_rearrange





import sys
import numpy as np
import pyqtgraph as pg
from PySide6.QtWidgets import QApplication
from pyqtgraph import Point
from PySide6 import QtGui, QtCore
from math import ceil, floor, frexp, isfinite, log10, sqrt
from pyqtgraph import debug
from pyqtgraph import functions as fn
import sys
import numpy as np
from PySide6 import QtWidgets, QtCore, QtGui
import pyqtgraph.opengl as gl
from OpenGL import GL
import enum
from pprint import pprint



class CustomItemGroup(pg.ItemGroup):
    def __init__(self, name):
        self.name=name
        pg.ItemGroup.__init__(self)

class CustomGraphicsObject(QtWidgets.QGraphicsObject):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.__flags = QtWidgets.QGraphicsItem.GraphicsItemFlag()
    
    def itemChange(self, change, value):

        res = super().itemChange(change, value)
        return res
    
    def setFlag(self, flag, enable):
        
        res = super().setFlag(flag, enable)
        
    
    def setFlag2(self, flag: enum.Flag, enabled=True):
        self.itemChange(self.GraphicsItemChange.ItemFlagsChange, enabled)
        flag_int = flag.value if hasattr(flag, 'value') else flag
        
        if enabled:
            new_flags = self.flags() | flag
        else:
            new_flags = self.flags() & ~flag
        
        if self.__flags != new_flags:
            self.__flags = new_flags
        self.itemChange(self.GraphicsItemChange.ItemFlagsHaveChanged, enabled)
        
    def flags2(self):
        return self.__flags
    
class CustomPlotDataItem(pg.PlotDataItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        
    def itemChange(self, change, value):
        res = super().itemChange(change, value)
        return res



class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.plot_item = pg.PlotItem()

        
    #    self.plot_item.setAutoVisible(True, True)

    #    self.legendItem =pg.LegendItem((80,60), offset=(70,20))
      #  self.legendItem.setParentItem(self.plot_item)
        
        

        """
        plot_item.setAxisItems({"left": pg.AxisItem(orientation="left"),
                                "bottom": pg.AxisItem(orientation="bottom")
                                }
                               )
        """
        

        
        ax1 = self.plot_item.getAxis("left")
        ax1.enableAutoSIPrefix(True)
      #  ax1.setStyle(tickLength=20)
        
        ax1.setLabel("abc")
        
        
        ax2 = self.plot_item.getAxis("bottom")
        ax2.enableAutoSIPrefix(True)

     #   ax2.setStyle(tickLength=20)
        
        
        x = np.linspace(0, 1, 20)
        y = 20*x
        
        self.plot_data = pg.PlotDataItem(x=x, y=y, pen='r', name="abcde")
        self.plot_item.addItem(self.plot_data)
        self.plot_data.setClipToView(True)
        self.plot_data.setDownsampling(None, True, "subsample")
        self.plot_data.setDynamicRangeLimit(1e5)
        self.plot_data.setSymbol("x")
    
        

        self.plot_data.update()

        self.timer = QtCore.QTimer()
        
      #  self.plot_item.setScale(100)
        
      
        self.timer.timeout.connect(self.set)
       # import sys
       # sys.exit()

        
         

        plot_widget = pg.PlotWidget(plotItem=self.plot_item)
        self.setCentralWidget(plot_widget)
        self.show()
  
        y*=2

        self.plot_data.setData(x, y)
        self.count=1
        #self.plot_data.setFlag(QtWidgets.QGraphicsItem.GraphicsItemFlag.ItemHasNoContents, False)
        

        
        
        
        

        # Run the Qt application
    def set(self):

        xdata= self.plot_data.xData
        xr = xdata.max() - xdata.min()
        
        xnew = np.linspace(xdata.min() - 0.1 * xr, xdata.max() + 0.1 * xr, xdata.size + 1)
        ynew_pre = self.plot_data.yData
        ynew = np.linspace(ynew_pre.min(), ynew_pre.max(), xnew.size)
        #self.plot_data.setData(xnew, ynew)
        self.plot_data.setData(y=1000 * self.plot_data.yData)
       # self.update()
        print(f"self.plot_data.scale: {self.plot_data.scale()}")
        
        #self.plot_item.update()
        self.count+=1

def main():
    """Main function to run the application."""
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    window.timer.start(1000)
    sys.exit(app.exec())

if __name__ == "__main__":
    main()


from PySide6.QtCore import QTimer, QThread, Signal, QObject
import threading
from PySide6.QtWidgets import QApplication, QWidget, QPushButton, QHBoxLayout, QVBoxLayout
import sys


class ModbusComWorker(QObject):
    finished = Signal()
    count=0
    def __init__(self, parent=None):
        self._thread = QThread()
        super().__init__(parent)
        self.moveToThread(self._thread)


        
        self._thread.started.connect(self.start)
        self._thread.start()



    def start(self):
        print(f"creating thread: {QThread.currentThread()}")
        self._timer = QTimer(self)
        self._timer.timeout.connect(self.process)
        self._timer.start(2000)

    def stop(self):
        self._timer.stop()
        self.finished.emit()

    def process(self):
        print(QThread.currentThread())
        self.count+=1
        if self.count==3:
            
            sys.exit()

class Widget(QWidget):
    def __init__(self):
        super().__init__()
        self._workers=[]
        self.button = QPushButton("Start Worker")
        self.button.clicked.connect(self.spawn_worker)
        layout = QVBoxLayout()
        layout.addWidget(self.button)
        self.setLayout(layout)
        self._threads = []
        self.worker_count = 0
        self.show()
        
    def spawn_worker(self):
        self.worker = ModbusComWorker()

if __name__ == "__main__":

    app = QApplication()
    if app is None:
        app = QApplication(sys.argv)

    
    w = Widget()

    print(F"starting: {QThread.currentThread()}")

    sys.exit(app.exec())


from __future__ import annotations
from typing import List, Dict, Union, Tuple, Callable, TYPE_CHECKING
if TYPE_CHECKING:
    from pyvol_terminal.windows import InterfaceWindow, SubWindow
    from instruments.utils import InstrumentManager

from PySide6 import QtCore, QtWidgets
from . import windows
from pyvol_terminal import workers
from pyvol_terminal.data_classes import builders as builders_data_classes

window_container: Dict[str, InterfaceWindow|SubWindow] = {}


class Application(QtWidgets.QApplication):
    create_window_signal = QtCore.Signal(dict, str, tuple)
    window_titles_signal = QtCore.Signal(list)
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance
    
    @classmethod
    def get_instance(cls):
        return cls._instance
    
    def __init__(self,
                 argv: List[str],
                 **window_configs):
        super().__init__(argv)
        self.window_configs=window_configs
        self._initialized = True
        self.window_titles_opened=[]
        self.n_windows=0
        self._worker_threads = {}
        
        
        self._price_process_worker = workers.PriceProcessor(window_configs["instrument_manager"],
                                                            data_processing_config=window_configs["data_processing_config"],
                                                            ws_config=window_configs["websocket_config"]
                                                            )

        self._websocket_worker = workers.WebsocketWorker()
        self._websocket_worker.update_signal.connect(self._price_process_worker.update)
        self.instrument_manager: InstrumentManager=window_configs["instrument_manager"]
        self._price_types = list(window_configs["raw_options_container"].keys())
        
        self.surface_engine_worker=workers.SurfaceCalibration(**self.window_configs, **self.window_configs["surface_construction_config"])
        self._price_process_worker.processedSig.connect(self.surface_engine_worker.price_process_response)
        
    def start(self):       
        self.open_main_window()    
        self._websocket_worker.start()    
        self.surface_engine_worker.start()
        #self.aboutToQuit.connect(self._cleanup)
    
    @QtCore.Slot()
    def price_processed(self):
        pass
    
    def calibrate_surface(self):
        pass
        
    def open_main_window(self):
        for i in range(1, 9999):
            window_title = f"Main Window {i}"
            if not window_title in self.window_titles_opened:
                break 
            
        vol_vect_container = builders_data_classes.create_vol_vect_container(self.surface_engine_worker.raw_options_container,
                                                                             self.instrument_manager.options_instrument_container.original_data,
                                                                            )
        self.window_configs["vol_vect_container"]=vol_vect_container
        win = windows.InterfaceWindow(self,
                                 **self.window_configs
                                 )
        
        window_container[window_title] = win
        
        self.window_titles_opened.append(window_title)    
        self.surface_engine_worker.add_vol_vect_container(window_title,
                                                          vol_vect_container
                                                          )
        win.show()

    def open_sub_window(self, widget: QtWidgets.QWidget, geometry: QtCore.Qsize, parent: QtWidgets.QWidget) -> windows.SubWindow:
        for i in range(1, 9999):
            window_title = f"Sub Window {i}"
            if not window_title in self.window_titles_opened:
                break 
        win = windows.SubWindow(widget, geometry, parent=parent)
        window_container[window_title] = win
        self.window_titles_opened.append(window_title)


    def _init_workers(self, configs):
        self.instrument_manager: InstrumentManager=configs["instrument_manager"]
        """Initialize all workers in their own threads"""
        # Price Processor Worker
        self._price_process_worker = workers.PriceProcessor(
            configs["instrument_manager"],
            data_processing_config=configs["data_processing_config"]
        )
        price_thread = QtCore.QThread()
        self._worker_threads['price'] = price_thread
        self._price_process_worker.moveToThread(price_thread)
        price_thread.start()
        
        # Websocket Worker
        self._websocket_worker = workers.WebsocketWorker(**configs["websocket_config"])
        ws_thread = QtCore.QThread()
        self._worker_threads['websocket'] = ws_thread
        self._websocket_worker.moveToThread(ws_thread)
        ws_thread.start()
        
        # Surface Engine Worker
        self.surface_engine_worker = workers.SurfaceCalibration(
            **self.window_configs,
            **self.window_configs["surface_construction_config"]
        )
        surface_thread = QtCore.QThread()
        self._worker_threads['surface'] = surface_thread
        self.surface_engine_worker.moveToThread(surface_thread)
        surface_thread.start()
        
        # Connect signals
        self._websocket_worker.update_signal.connect(self._price_process_worker.update)
        self._price_process_worker.processedSig.connect(self.surface_engine_worker.price_process_response)
        
    def _cleanup(self):
        """Clean up all worker threads"""
        for name, thread in self._worker_threads.items():
            worker = getattr(self, f'_{name}_worker', None) or getattr(self, f'{name}_worker', None)
            if worker:
                worker.stop()  # Ensure all workers have a stop method
            thread.quit()
            thread.wait()


#%%
from enum import Flag, auto

class ItemChangeFlag(Flag):
    """Custom flags for item changes (like Qt's GraphicsItemChange)."""
    ITEM_POSITION_CHANGE = auto()      # 1
    ITEM_VISIBLE_CHANGE = auto()       # 2
    ITEM_SELECTION_CHANGE = auto()     # 4
    ITEM_PARENT_CHANGE = auto()        # 8
    ITEM_TRANSFORM_CHANGE = auto()     # 16
    
    
class MyCustomItem:
    def __init__(self):
        self._position = (0, 0)
        self._visible = True

    def itemChange(self, change: ItemChangeFlag, value):
        """Mimics QGraphicsItem.itemChange()."""
        if change == ItemChangeFlag.ITEM_POSITION_CHANGE:
            print(f"Position changed to: {value}")
            self._position = value
        elif change == ItemChangeFlag.ITEM_VISIBLE_CHANGE:
            print(f"Visibility changed to: {value}")
            self._visible = value
        elif change in (ItemChangeFlag.ITEM_SELECTION_CHANGE | ItemChangeFlag.ITEM_PARENT_CHANGE):
            print(f"Combined change detected: {change}")
        else:
            print(f"Unhandled change: {change}")

        return value  # (Optional: Return modified value, like Qt)
    
item = MyCustomItem()

# Single flag (position change)
item.itemChange(ItemChangeFlag.ITEM_POSITION_CHANGE, (100, 200))


combined_flags = ItemChangeFlag.ITEM_SELECTION_CHANGE | ItemChangeFlag.ITEM_PARENT_CHANGE
item.itemChange(combined_flags, None)

# Check if a flag is set in a combined value
flags = ItemChangeFlag.ITEM_VISIBLE_CHANGE | ItemChangeFlag.ITEM_TRANSFORM_CHANGE
if ItemChangeFlag.ITEM_VISIBLE_CHANGE in flags:
    print("Visibility flag is set!")  # This runs
    
    
#%%%

from enum import Flag, auto

class ItemChangeFlag(Flag):
    """Flags representing different item change types"""
    ITEM_VISIBLE_CHANGE = auto()
    ITEM_POSITION_CHANGE = auto()
    ITEM_SELECTION_CHANGE = auto()
    ITEM_PARENT_CHANGE = auto()
    ITEM_TRANSFORM_CHANGE = auto()

class GraphicsItem:
    def __init__(self, parent=None):
        self.visible = True
        self.explicitly_hidden = False
        self.parent = parent  # Reference to parent item

    def set_visible_helper(self, new_visible, explicitly=False, update=True, hidden_by_panel=False):
        """
        Mimics QGraphicsItemPrivate::setVisibleHelper
        - new_visible: Target visibility state
        - explicitly: Whether this is an explicit user-set visibility change
        - update: Whether to trigger visual updates
        - hidden_by_panel: Special case for panel visibility
        """
        # Update explicit hidden flag
        if explicitly:
            self.explicitly_hidden = not new_visible

        # Check if visibility is already in target state
        if self.visible == new_visible:
            return

        # Don't show if parent exists and is hidden
        if self.parent and new_visible and not self.parent.visible:
            return

        # Call itemChange to potentially modify the visibility
        modified_visible = self.item_change(
            ItemChangeFlag.ITEM_VISIBLE_CHANGE,
            new_visible
        )

        # Check if itemChange modified the visibility
        if self.visible == modified_visible:
            return

        # Update visibility state
        self.visible = modified_visible
        
        # Additional logic would go here (update children, emit signals, etc.)
        if update:
            self.update_visuals()
    
    def item_change(self, change, value):
        """
        Customizable change handler - similar to QGraphicsItem.itemChange
        Override this in subclasses to modify behavior
        """
        # Base implementation just returns the value unchanged
        return value

    def update_visuals(self):
        """Placeholder for visual update logic"""
        print(f"Visibility updated to: {self.visible}")
        # Would typically trigger redraw/update here

# Example subclass with custom itemChange behavior
class CustomItem(GraphicsItem):
    def item_change(self, change, value):
        if change == ItemChangeFlag.ITEM_VISIBLE_CHANGE:
            print(f"Visibility change requested: {value}")
            # Add custom logic here - e.g., prevent hiding in some cases
            if value is False:
                print("Preventing item from being hidden!")
                return True  # Override to remain visible
            return value
        return super().item_change(change, value)

# Usage example
if __name__ == "__main__":
    parent_item = GraphicsItem()
    parent_item.visible = False  # Parent is hidden
    
    child_item = CustomItem(parent=parent_item)
    
    print("Trying to show child (should be blocked by parent):")
    child_item.set_visible_helper(True)
    # Output: (no change, blocked by parent)
    
    print("\nTrying to hide child with custom logic:")
    child_item.set_visible_helper(False)
    # Output: 
    # Visibility change requested: False
    # Preventing item from being hidden!
    # Visibility updated to: True


import abc
from PySide6.QtCore import QObject

class QABCMeta(abc.ABCMeta, type(QObject)):
    
    """
    SOURCE:
    https://stackoverflow.com/questions/78778926/creating-a-metaclass-that-inherits-from-abcmeta-and-qobject
    
    """
    
    def __new__(mcls, name, bases, ns, **kw):
        cls = super().__new__(mcls, name, bases, ns, **kw)
        abc._abc_init(cls)
        return cls
    def __call__(cls, *args, **kw):
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        return super().__call__(*args, **kw)


#%%


from PySide6 import QtCore, QtWidgets
from PySide6.QtWidgets import QApplication
from abc import abstractclassmethod, abstractmethod

import abc
from PySide6.QtCore import QObject

class QABCMeta(abc.ABCMeta, type(QObject)):
    
    """
    SOURCE:
    https://stackoverflow.com/questions/78778926/creating-a-metaclass-that-inherits-from-abcmeta-and-qobject
    
    """
    
    def __new__(mcls, name, bases, ns, **kw):
        cls = super().__new__(mcls, name, bases, ns, **kw)
        abc._abc_init(cls)
        return cls
    def __call__(cls, *args, **kw):
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        return super().__call__(*args, **kw)



class ABCInterface(QtWidgets.QWidget, metaclass=QABCMeta):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.on_view=False
        self.main_view: _MainViewMeta=None
        self.hide()
        

